from .deletion import deletion
from ._insertion import insertion
from ._irof import irof
from ._impact_coverage import impact_coverage
from .infidelity import infidelity
from ._max_sensitivity import max_sensitivity
from .minimal_subset import minimal_subset
from .sensitivity_n import sensitivity_n
from ._parameter_randomization import parameter_randomization


import numpy as np
from numpy import typing as npt
from scipy import stats


def wilcoxon_tests(df, higher_is_better):
    pvalues, effect_sizes = {}, {}
    for method_name in df:
        method_results = df[method_name].to_numpy()
        _, pvalue = stats.wilcoxon(
            method_results,
            alternative="greater" if higher_is_better else "less",
        )
        pvalues[method_name] = pvalue
        effect_sizes[method_name] = np.median(method_results)
    return effect_sizes, pvalues


def rowwise_pearsonr(a: npt.NDArray, b: npt.NDArray) -> npt.NDArray:
    """
    Calculates row-wise correlations between two arrays.
    This is a faster implementation of scipy.stats.pearsonr,
    as it only calculates correlation coefficients between corresponding rows,
    rather than between all pairs of rows.

    Parameters
    ----------
    a : npt.NDArray
        first set of row vectors (shape: [num_rows, num_measurements])
    b : npt.NDArray
        second set of row vectors (shape: [num_rows, num_measurements])

    Returns
    -------
    npt.NDArray
        row-wise correlations between a and b (shape: [num_rows])
    """
    # Subtract mean
    # [batch_size, num_observations]
    a -= a.mean(axis=1, keepdims=True)
    b -= b.mean(axis=1, keepdims=True)
    # Calculate numerator
    # [batch_size]
    cov = (a * b).sum(axis=1)
    # Calculate denominator
    # [batch_size]
    denom = np.sqrt((a**2).sum(axis=1)) * np.sqrt((b**2).sum(axis=1))
    # denom_zero = denom == 0.0
    # if np.any(denom_zero):
    #    warnings.warn(f"Zero standard deviation detected")

    # If the denominator is zero, that means one of the series is constant.
    # Correlation is technically undefined in this case, but covariance is 0.
    # We can just set the correlation coefficient to zero in this case.
    corrcoefs = np.divide(cov, denom, out=np.zeros_like(cov), where=denom != 0)
    return corrcoefs


def rowwise_spearmanr(a: npt.NDArray, b: npt.NDArray) -> npt.NDArray:
    """
    Calculates row-wise Spearman correlations between two arrays.
    This is a faster implementation of scipy.stats.spearmanr,
    as it only calculates correlation coefficients between corresponding rows,
    rather than between all pairs of rows.

    Parameters
    ----------
    a : npt.NDArray
        first set of row vectors (shape: [num_rows, num_measurements])
    b : npt.NDArray
        second set of row vectors (shape: [num_rows, num_measurements])

    Returns
    -------
    npt.NDArray
        row-wise Spearman correlations between a and b (shape: [num_rows])
    """
    # Calculate rank of each row
    # [batch_size, num_observations]
    a_ranks = stats.rankdata(a, axis=1)
    b_ranks = stats.rankdata(b, axis=1)

    # Spearman rank correlation is simply Pearson correlation of ranks
    return rowwise_pearsonr(a_ranks, b_ranks)


from typing import Tuple, Dict, Union, Optional, List
import os
import yaml
import numpy as np
from numpy import typing as npt
import h5py


class RandomAccessNDArrayTree:
    def __init__(
        self,
        levels: Dict[str, List[str]],
        shape: List[int],
    ):
        """
        Represents a tree of numpy NDArrays capable of writing in a
        random-access way (as opposed to append-only).
        :param levels: represents the names and keys at each level of the tree.
            E.g: {"masker": ["constant", "random"], "method": ["m1", "m2"]}
            will create a tree with 2 levels: a masker level and method level.
            Each combination of masker and method corresponds to 1 NDArray.
        """
        self.levels = levels
        self.shape = shape
        self.level_names: List[str] = sorted(list(levels.keys()))

        def _initialize_data(data, depth=0) -> Dict:
            """
            Recursively initialize the data structure.
            If we are in an internal node, a dictionary is created and keys for
            the next level are added recursively. Otherwise, a 0-valued NDArray
            of the necessary shape is created.
            :param data: The nested dictionary with the data we are constructing
            :param depth: The current depth
            :return: Nested dictionary containing the tree of NDArrays
            """
            level_keys = levels[self.level_names[depth]]
            for key in level_keys:
                if depth < len(levels) - 1:
                    data[key] = {}
                    _initialize_data(data[key], depth + 1)
                else:
                    data[key] = np.zeros(shape=self.shape)
            return data

        self._data: Dict = _initialize_data({})

    def write_dict(
        self,
        indices: npt.NDArray,
        data: Dict,
        level_order: Optional[List[str]] = None,
    ):
        """
        Recursively write a dictionary of data to fixed indices.
        The data dictionary has to follow the same levels structure as the tree.
        For each combination of levels, data is written to the same indices.
        This is analogous to recursively looping over all keys in data and
        calling tree.write(indices, data[level1][level2]) for each NDArray.
        :param indices: Indices where data should be written for each NDArray
        :param data: Dictionary containing the data to be written
        :param level_order: List of strings indicating in which order the
            different levels are encountered in the dictionary.
            If not specified, alphabetical order is used.
        """
        if level_order is None:
            level_order = self.level_names
        elif sorted(level_order) != sorted(self.level_names):
            raise ValueError(f"Invalid level_order: {level_order}")

        def _write_rec(_data, level_keys):
            depth = len(level_keys.keys())
            if depth < len(self.level_names):
                cur_level_name = level_order[depth]
                for key in _data.keys():
                    # _data is a dictionary, loop over all keys and recurse
                    level_keys[cur_level_name] = key
                    _write_rec(_data[key], dict(level_keys))
            else:
                # We have reached a leaf, _data is an NDArray.
                # Write to the specified location.
                self.write(indices, _data, **level_keys)

        _write_rec(data, {})

    def write_dict_split(
        self,
        indices_dict: Dict[str, npt.NDArray],
        target_indices: npt.NDArray,
        split_level: str,
        data: Dict,
        level_order: Optional[List[str]] = None,
    ):
        """
        Same as write_dict, but instead of writing the full NDArray to each
        location, indices in each NDArray are chosen based on some level.
        For example: write indices [1,2,4] to all locations with
        method_name == "Gradient", and indices [3, 5] to all locations with
        method_name == "InputXGradient".
        :param indices_dict: Dictionary mapping each level key of the split
            level to its corresponding indices in the original data.
            Do not confuse with indices in write or write_dict: these are
            target indices (indices where the data should be written to).
        :param target_indices: NDArray containing the indices where the data
            should be written to.
        :param split_level: The level at which the NDArrays must be split.
            This level is not present in data.
        :param data: Dictionary containing the data to be written
        :param level_order: List of strings indicating in which order the
            different levels are encountered in the dictionary.
            If not specified, alphabetical order is used.
        """
        # This contains the levels that should be present in the data dict
        # and in level_order. These are all levels except the split level.
        data_levels = list(self.level_names)
        data_levels.remove(split_level)

        if level_order is None:
            level_order = data_levels
        elif sorted(level_order) != sorted(data_levels):
            raise ValueError(f"Invalid level_order: {level_order}")

        def _write_rec(_data, level_keys, split_key):
            depth = len(level_keys.keys())
            if depth < len(self.level_names):
                # We index using depth-1 because level_order does not contain
                # the split level.
                cur_level_name = level_order[depth - 1]
                for key in _data.keys():
                    # Loop over all keys and recurse
                    level_keys[cur_level_name] = key
                    _write_rec(_data[key], dict(level_keys), split_key)
            else:
                # We have reached a leaf, _data is an NDArray.
                # Write to the specified location.
                cur_indices = indices_dict[split_key]
                cur_target_indices = target_indices[cur_indices]
                self.write(
                    cur_target_indices, _data[cur_indices], **level_keys
                )

        # Write separately for each split key
        for key in indices_dict.keys():
            # We basically start at depth 1, the split_level is already set.
            # This level will not be present in the data dictionary.
            _write_rec(data, {split_level: key}, split_key=key)

    def write(
        self, indices: npt.NDArray, data: npt.NDArray, **level_keys
    ) -> None:
        """
        Writes data to the given indices (axis=0) of a certain NDArray.
        The NDArray is given by the level_keys: for each level name there
        should be 1 key. This encodes the path to take down the tree.
        :param indices: Indices of NDArray where data should be written
        :param data: The data to write in the NDArray at the given indices
        :param level_keys: a key for each level to indicate the path to the
            desired NDArray
        """
        if set(level_keys.keys()) != set(self.levels.keys()):
            raise ValueError(
                f"Must provide key for all levels: {list(self.levels.keys())}"
                "(received {level_keys})"
            )
        dest: Union[Dict, npt.NDArray] = self._data
        # Find the relevant NDArray
        for level_name in self.level_names:
            dest = dest[level_keys[level_name]]
        # At this point, dest should be an NDArray. Write the data.
        dest[indices] = data

    def get(self, **level_keys) -> npt.NDArray:
        """
        Get a single NDArray from the tree
        :param level_keys: Keys encoding the path to take down the tree
        """
        if set(level_keys.keys()) != set(self.levels.keys()):
            raise ValueError(
                f"Must provide key for all levels: {list(self.levels.keys())}"
            )
        res = self._data
        for level_name in self.level_names:
            res = res[level_keys[level_name]]
        if not isinstance(res, np.ndarray):
            raise ValueError(
                f"Expected leaf node to be ndarray, got {type(res)}"
            )
        return res

    def save_to_hdf(self, group: h5py.Group) -> None:
        """
        Attach the NDArrayTree to an HDF5 group
        :param group: the HDF5 group to which the root of this NDArrayTree
            should be attached
        """

        def _add_rec(cur_data, cur_group, depth=0):
            for key in cur_data:
                cur_group.attrs["level_name"] = self.level_names[depth]
                if depth < len(self.level_names) - 1:
                    next_group = cur_group.create_group(key)
                    _add_rec(cur_data[key], next_group, depth + 1)
                else:
                    cur_group.create_dataset(key, data=cur_data[key])

        _add_rec(self._data, group)

    @classmethod
    def load_from_hdf(cls, group: h5py.Group) -> "RandomAccessNDArrayTree":
        def _load_metadata_rec(
            node: Union[h5py.Group, h5py.Dataset], cur_levels
        ) -> Tuple[Dict, List[int]]:
            """
            Recursively loads the necessary metadata from the tree:
            level names, level keys, NDArray shape.
            :return: the necessary arguments to instantiate an NDArrayTree from
                this file.
            """
            if isinstance(node, h5py.Group):
                keys = list(node.keys())
                level_name = node.attrs["level_name"]
                cur_levels[level_name] = keys
                next_node = node[keys[0]]
                assert isinstance(next_node, (h5py.Group, h5py.Dataset))
                return _load_metadata_rec(next_node, cur_levels)
            elif isinstance(node, h5py.Dataset):
                return cur_levels, list(node.shape)

        def _copy_ndarrays_rec(h5py_node: h5py.Group, tree_node: Dict):
            """
            Recursively copy the NDArrays from the HDF5 file to the tree
            """
            for key in list(h5py_node.keys()):
                next_node = h5py_node[key]
                if isinstance(next_node, h5py.Dataset):
                    # We are at the bottom, copy the datasets to the tree
                    # An HDF5 Dataset can be converted to an NDArray using [()]
                    tree_node[key] = next_node[()]
                elif isinstance(next_node, h5py.Group):
                    # We are at an inner node, descend
                    _copy_ndarrays_rec(next_node, tree_node[key])
                else:
                    raise ValueError(f"Invalid type: {type(next_node)}")

        # Instantiate the NDArrayTree with the necessary metadata
        levels, shape = _load_metadata_rec(group, {})
        result = cls(levels, shape)

        # Traverse the NDArrayTree to copy the NDArrays to their respective
        # locations
        _copy_ndarrays_rec(group, result._data)
        return result

    def save_to_dir(self, path: str) -> None:
        """
        Save the NDArrayTree to a directory of CSV files with metadata in
        a yaml file.
        :param path: The path to the directory
        """
        # First read any existing metadata
        with open(f"{path}/metadata.yaml", "r") as fp:
            metadata = yaml.safe_load(fp) or {}

        # Add the necessary metadata
        metadata["level_names"] = self.level_names
        metadata["levels"] = self.levels
        metadata["shape"] = self.shape

        # Write the metadata
        with open(f"{path}/metadata.yaml", "w") as fp:
            yaml.safe_dump(metadata, fp)

        # Write the NDArrays
        def _add_rec(cur_data, cur_path, depth=0):
            for key in cur_data:
                if depth < len(self.level_names) - 1:
                    next_path = os.path.join(cur_path, key)
                    _add_rec(cur_data[key], next_path, depth + 1)
                else:
                    if not os.path.exists(cur_path):
                        os.makedirs(cur_path, exist_ok=True)
                    np.savetxt(
                        os.path.join(cur_path, f"{key}.csv"), cur_data[key]
                    )

        _add_rec(self._data, path)

    @classmethod
    def load_from_dir(cls, path: str) -> "RandomAccessNDArrayTree":
        # Load metadata
        with open(os.path.join(path, "metadata.yaml"), "r") as fp:
            metadata = yaml.safe_load(fp)

        # Instantiate the NDArrayTree with the necessary metadata
        result = cls(metadata["levels"], metadata["shape"])

        # Traverse the NDArrayTree to copy the NDArrays to their respective
        # locations
        def _copy_ndarrays_rec(cur_path, tree_node):
            """
            Recursively copy the NDArrays from the directory to the tree
            """
            for key in os.listdir(cur_path):
                next_path = os.path.join(cur_path, key)
                if os.path.isdir(next_path):
                    # We are at an inner node, descend
                    _copy_ndarrays_rec(next_path, tree_node[key])
                elif key.endswith(".csv"):
                    # We are at the bottom, copy the datasets to the tree
                    tree_node[key[:-4]] = np.loadtxt(next_path)

        _copy_ndarrays_rec(path, result._data)
        return result



from typing import List, Mapping
import torch
from attribench.masking.image import ImageMasker
from attribench._model_factory import ModelFactory
from ..deletion._deletion_worker import DeletionWorker
from .._metric_worker import WorkerConfig
from attribench.data import AttributionsDataset
from attribench.functional.metrics._irof import _irof_batch


class IrofWorker(DeletionWorker):
    def __init__(
        self,
        worker_config: WorkerConfig,
        model_factory: ModelFactory,
        dataset: AttributionsDataset,
        batch_size: int,
        maskers: Mapping[str, ImageMasker],
        activation_fns: List[str],
        mode: str = "morf",
        start: float = 0,
        stop: float = 1,
        num_steps: int = 100,
    ):
        super().__init__(
            worker_config,
            model_factory,
            dataset,
            batch_size,
            maskers,
            activation_fns,
            mode,
            start,
            stop,
            num_steps,
        )
        self.maskers = maskers

    def process_batch(
        self,
        batch_x: torch.Tensor,
        batch_y: torch.Tensor,
        batch_attr: torch.Tensor,
    ):
        return _irof_batch(
            batch_x,
            batch_y,
            self.model,
            batch_attr,
            self.maskers,
            self.activation_fns,
            self.mode,
            self.start,
            self.stop,
            self.num_steps,
        )



from abc import abstractmethod
from typing import Tuple
import numpy as np
import torch


class Masker:
    """Base class for all maskers.
    Maskers are used to "remove" features from a sample by masking them with
    some value. This can be a fixed baseline value, a random value, or some
    other value.

    Note that a Masker object is not yet usable after creation. You need to
    call :meth:`set_batch` first, to set the samples and attributions.
    This allows the same Masker object to be used for multiple batches.
    """

    def __init__(self):
        self.baseline: torch.Tensor | None = None
        self.samples: torch.Tensor | None = None
        self.attributions: torch.Tensor | None = None
        self.sorted_indices: torch.Tensor | None = None
        self.rng = np.random.default_rng()

    def get_num_features(self) -> int:
        """Return the number of features in the samples.

        Returns
        -------
        int
            Number of features in the samples.
        """
        assert self.sorted_indices is not None
        return self.sorted_indices.shape[1]

    def mask_top(self, k: int) -> torch.Tensor:
        """Mask the ``k`` most important features, according to the attributions.

        Parameters
        ----------
        k : int
            Number of features to mask.

        Returns
        -------
        torch.Tensor
            Samples with the top k features masked.
        """
        assert self.sorted_indices is not None
        assert self.samples is not None
        if k == 0:
            return self.samples
        else:
            return self._mask(self.sorted_indices[:, -k:])

    def mask_bot(self, k: int) -> torch.Tensor:
        """Mask the ``k`` least important features, according to the attributions.

        Parameters
        ----------
        k : int
            Number of features to mask.

        Returns
        -------
        torch.Tensor
            Samples with the bottom k features masked.
        """
        assert self.sorted_indices is not None
        return self._mask(self.sorted_indices[:, :k])

    def mask_rand(
        self, k: int, return_indices=False
    ) -> torch.Tensor | Tuple[torch.Tensor, torch.Tensor]:
        """Mask ``k`` random features.

        Parameters
        ----------
        k : int
            Number of features to mask.
        return_indices : bool, optional
            Whether to return the indices of the masked features, by default False

        Returns
        -------
        torch.Tensor
            Samples with k random features masked.
        """
        assert self.samples is not None
        if k == 0:
            return self.samples

        num_samples = self.samples.shape[0]
        num_features = self.get_num_features()

        indices = torch.tensor(np.tile(
            self.rng.choice(num_features, size=k, replace=False),
            (num_samples, 1),
        ))
        masked_samples = self._mask(indices)
        if return_indices:
            return masked_samples, indices
        return masked_samples

    @abstractmethod
    def set_batch(
        self, samples: torch.Tensor, attributions: torch.Tensor | None = None
    ):
        """Set the samples and attributions for the next batch.

        Parameters
        ----------
        samples : torch.Tensor
            Samples of shape ``[num_samples, *sample_shape]``.
        attributions : torch.Tensor, optional
            Attributions of shape ``[num_samples, *sample_shape]``, by default None
            If None, the :meth:`mask_top` and :meth:`mask_bot` methods will 
            not be available.
        """
        raise NotImplementedError

    @abstractmethod
    def _check_attribution_shape(self, samples, attributions):
        """Check if the attributions have the correct shape."""
        raise NotImplementedError

    @abstractmethod
    def _mask(self, indices: torch.Tensor) -> torch.Tensor:
        """Mask the given indices in the samples."""
        raise NotImplementedError

    @abstractmethod
    def _mask_boolean(self, bool_mask: torch.Tensor) -> torch.Tensor:
        """Mask using the given boolean mask."""
        raise NotImplementedError



from numpy import typing as npt
import torch
from .._metric_worker import MetricWorker
from ..._worker import WorkerConfig
from typing import Callable, Dict

from torch import nn

from attribench.masking import Masker
from attribench.data import AttributionsDataset
from attribench.functional.metrics.minimal_subset._minimal_subset import (
    minimal_subset_batch,
)


class MinimalSubsetWorker(MetricWorker):
    def __init__(
        self,
        worker_config: WorkerConfig,
        model_factory: Callable[[], nn.Module],
        dataset: AttributionsDataset,
        batch_size: int,
        maskers: Dict[str, Masker],
        mode: str,
        num_steps: int,
    ):
        super().__init__(worker_config, model_factory, dataset, batch_size)
        self.maskers = maskers
        self.mode = mode
        self.num_steps = num_steps

    def process_batch(
        self,
        batch_x: torch.Tensor,
        batch_y: torch.Tensor,
        batch_attr: torch.Tensor,
    ):
        return minimal_subset_batch(
            batch_x,
            self.model,
            batch_attr,
            self.num_steps,
            self.maskers,
            self.mode,
        )



from abc import abstractmethod
from torch import multiprocessing as mp
from torch import distributed as dist
from ._message import PartialResultMessage, DoneMessage
from typing import Callable
from multiprocessing.synchronize import Event


class WorkerConfig:
    def __init__(self, world_size: int, rank: int) -> None:
        self.world_size = world_size
        self.rank = rank

    @abstractmethod
    def setup(self):
        raise NotImplementedError

    @abstractmethod
    def cleanup(self):
        raise NotImplementedError

    @abstractmethod
    def send_result(self, result: PartialResultMessage):
        raise NotImplementedError


class DistributedWorkerConfig(WorkerConfig):
    def __init__(
        self,
        world_size: int,
        result_queue: mp.Queue,
        rank: int,
        all_processes_done: Event,
    ):
        super().__init__(world_size, rank)
        self.result_queue = result_queue
        self.all_processes_done = all_processes_done

    def setup(self):
        dist.init_process_group(
            "gloo", rank=self.rank, world_size=self.world_size
        )

    def cleanup(self):
        self.result_queue.put(DoneMessage(self.rank))
        self.all_processes_done.wait()
        dist.destroy_process_group()

    def send_result(self, result: PartialResultMessage):
        self.result_queue.put(result)


class SynchronizedWorkerConfig(WorkerConfig):
    def __init__(self, result_handler: Callable[[PartialResultMessage], None]):
        super().__init__(1, 0)
        self.result_handler = result_handler

    def setup(self):
        pass

    def cleanup(self):
        pass

    def send_result(self, result: PartialResultMessage):
        self.result_handler(result)


class Worker:
    def __init__(
        self,
        worker_config: WorkerConfig,
    ):
        self.worker_config = worker_config

    def run(self):
        self.worker_config.setup()
        self.work()
        self.worker_config.cleanup()

    def work(self):
        raise NotImplementedError



from .._metric import Metric
from .._metric_worker import MetricWorker
from ..._worker import WorkerConfig
from attribench.result import ImpactCoverageResult
from typing import Tuple, Optional
from torch.utils.data import Dataset
from attribench.data import IndexDataset
from attribench._method_factory import MethodFactory
from attribench._model_factory import ModelFactory
from ._impact_coverage_worker import ImpactCoverageWorker


class ImpactCoverage(Metric):
    """Computes the Impact Coverage metric for a given dataset, model, and
    set of attribution methods, using multiple processes.

    Impact Coverage is computed by applying an adversarial patch to the input.
    This patch causes the model to change its prediction.
    The Impact Coverage metric is the intersection over union (IoU) of the
    patch with the top n attributions of the input, where n is the number of
    features masked by the patch. The idea is that, as the patch causes the
    model to change its prediction, the corresponding region in the image
    should be highly relevant to the model's prediction.

    Impact Coverage requires a folder containing adversarial patches. The
    patches should be named as follows: patch_<target>.pt, where <target>
    is the target class of the patch. The target class is the class that
    the model will predict when the patch is applied to the input.

    The number of processes is determined by the number of devices. If
    `devices` is None, then all available devices are used. Samples are
    distributed evenly across the processes.

    To generate adversarial patches, the 
    :meth:`~attribench.functional.train_adversarial_patches` function
    or :class:`~attribench.distributed.TrainAdversarialPatches` class
    can be used.
    """
    def __init__(
        self,
        model_factory: ModelFactory,
        samples_dataset: Dataset,
        batch_size: int,
        method_factory: MethodFactory,
        patch_folder: str,
        address="localhost",
        port="12355",
        devices: Optional[Tuple] = None,
    ):
        """
        Parameters
        ----------
        model_factory : ModelFactory
            ModelFactory instance or callable that returns a model.
            Used to create a model for each subprocess.
        dataset : Dataset
            Dataset to compute Impact Coverage for.
        batch_size : int
            Batch size to use when computing Impact Coverage.
        method_factory : MethodFactory
            MethodFactory instance or callable that returns a dictionary
            mapping method names to attribution methods, given a model.
        patch_folder : str
            Path to folder containing adversarial patches.
        address : str, optional
            Address to use for the multiprocessing connection,
            by default "localhost"
        port : str, optional
            Port to use for the multiprocessing connection,
            by default "12355"
        devices : Optional[Tuple], optional
            Devices to use. If None, then all available devices are used.
            By default None.
        """
        index_dataset = IndexDataset(samples_dataset)
        super().__init__(
            model_factory,
            index_dataset,
            batch_size,
            address,
            port,
            devices,
        )
        self.method_factory = method_factory
        self.patch_folder = patch_folder
        self._result = ImpactCoverageResult(
            method_factory.get_method_names(), len(index_dataset)
        )

    def _create_worker(
        self, worker_config: WorkerConfig
    ) -> MetricWorker:
        return ImpactCoverageWorker(
            worker_config,
            self.model_factory,
            self.dataset,
            self.batch_size,
            self.method_factory,
            self.patch_folder,
        )



import pandas as pd
import matplotlib.pyplot as plt
from typing import Dict, Tuple
import seaborn as sns
from matplotlib.figure import Figure
from attribench.plot import Plot


def _create_fig(df, figsize, annot):
    fig, ax = plt.subplots(figsize=figsize)
    sns.heatmap(
        df,
        annot=annot,
        vmin=-1,
        vmax=1,
        cmap=sns.diverging_palette(220, 20, as_cmap=True),
        ax=ax,
        fmt=".2f",
        cbar=False,
    )
    ax.set_aspect("equal")
    return fig, ax


class InterMetricCorrelationPlot(Plot):
    """Heatmap showing Spearman correlations between metrics."""

    def render(
        self,
        title: str | None = None,
        figsize: Tuple[int, int] = (20, 20),
        fontsize: int | None = None,
        annot: bool = False,
    ) -> Figure:
        """Render the plot.

        Parameters
        ----------
        title : str | None, optional
            Title of the figure, by default None
        figsize : Tuple[int, int], optional
            Size of the figure, by default (20, 20)
        fontsize : int | None, optional
            Font size of x and y axis ticks, by default None
        annot : bool, optional
            Whether to annotate the heatmap with the correlation values, by
            default False

        Returns
        -------
        Figure
            The rendered Matplotlib figure.
        """
        corr_dfs = []
        methods = list(self.dfs.values())[0][0].columns
        for method_name in methods:
            data = {}
            for metric_name, (df, higher_is_better) in self.dfs.items():
                data[metric_name] = (
                    -df[method_name].to_numpy()
                    if not higher_is_better
                    else df[method_name].to_numpy()
                )
            df = pd.DataFrame(data)
            corr_dfs.append(df.corr(method="spearman"))
        corr = pd.concat(corr_dfs).groupby(level=0).mean()
        corr = corr.reindex(corr.columns)
        fig, ax = _create_fig(corr, figsize, annot)
        if title is not None:
            ax.set_title(title)

        ax.set_xticklabels(
            ax.get_xticklabels(),
            rotation=45,
            ha="right",
            rotation_mode="anchor",
            fontsize=fontsize,
        )
        ax.set_yticklabels(ax.get_yticklabels(), fontsize=fontsize)
        return fig


class InterMethodCorrelationPlot(Plot):
    """Heatmap showing Spearman correlations between methods."""

    def render(
        self,
        title: str | None = None,
        figsize=(20, 20),
        fontsize: int | None = None,
        annot=False,
    ) -> Figure:
        """Render the plot.
        Spearman correlation values are averaged across metrics.
        To plot inter-method correlations for each metric separately,
        use :meth:`render_all`.

        Parameters
        ----------
        title : str | None, optional
            Title of the figure, by default None
        figsize : Tuple[int, int], optional
            Size of the figure, by default (20, 20)
        fontsize : int | None, optional
            Font size of x and y axis ticks, by default None
        annot : bool, optional
            Whether to annotate the heatmap with the correlation values, by
            default False

        Returns
        -------
        Figure
            The rendered Matplotlib figure.
        """
        # Compute correlations for each metric
        all_dfs = [
            df if not inverted else -df
            for _, (df, inverted) in self.dfs.items()
        ]
        corr_dfs = [df.corr(method="spearman") for df in all_dfs]

        # Compute average of correlations
        corr = pd.concat(corr_dfs).groupby(level=0).mean()
        fig, ax = _create_fig(corr, figsize, annot)
        if title is not None:
            ax.set_title(title)
        ax.set_xticklabels(
            ax.get_xticklabels(),
            rotation=45,
            ha="right",
            rotation_mode="anchor",
            fontsize=fontsize,
        )
        ax.set_yticklabels(ax.get_yticklabels(), fontsize=fontsize)
        return fig

    def render_all(
        self, figsize=(20, 20), fontsize: int | None = None, annot=False
    ) -> Dict[str, Figure]:
        """Render a separate heatmap for each metric.
        TODO test and make sure args are consistent with render.

        Parameters
        ----------
        figsize : Tuple[int, int], optional
            Size of the figures, by default (20, 20)
        fontsize : int | None, optional
            Font size of x and y axis ticks, by default None
        annot : bool, optional
            Whether to annotate the heatmaps with the correlation values, by
            default False

        Returns
        -------
        Dict[str, Figure]
            Dictionary mapping metric names to rendered Matplotlib figures.
        """
        figs = {}
        for name, (df, inverted) in self.dfs.items():
            if inverted:
                df = -df
            corr = df.corr(method="spearman")
            figs[name] = _create_fig(corr, figsize, annot)
        return figs



from attribench.result import MaxSensitivityResult
from attribench.result._grouped_batch_result import GroupedBatchResult
from typing import Dict
from attribench.data import AttributionsDataset
from attribench.data.attributions_dataset._attributions_dataset import (
    GroupedAttributionsDataset,
)
import torch
from attribench._attribution_method import AttributionMethod
from torch.utils.data import DataLoader
import math


def _normalize_attrs(attrs):
    flattened = attrs.flatten(1)
    return flattened / torch.norm(flattened, dim=1, p=math.inf, keepdim=True)


def _max_sensitivity_batch(
    batch_x: torch.Tensor,
    batch_y: torch.Tensor,
    batch_attr: Dict[str, torch.Tensor],
    method_dict: Dict[str, AttributionMethod],
    num_perturbations: int,
    radius: float,
    device: torch.device,
) -> Dict[str, torch.Tensor]:
    if set(method_dict.keys()) != set(batch_attr.keys()):
        print(method_dict.keys())
        print(batch_attr.keys())
        raise ValueError(
            "Method dictionary and batch attributions dictionary"
            " must have the same keys."
        )
    result: Dict[str, torch.Tensor] = {
        method_name: torch.zeros(1) for method_name in method_dict.keys()
    }
    batch_x = batch_x.to(device)
    batch_y = batch_y.to(device)

    # Compute Max-Sensitivity for each method
    for method_name, method in method_dict.items():
        #attrs = _normalize_attrs(batch_attr[method_name])
        # TODO we are not taking into account the aggregation here
        # TODO also the AttributionsDataset is not being used
        attrs = _normalize_attrs(method(batch_x, batch_y).detach()).cpu()
        diffs = []

        for _ in range(num_perturbations):
            # Add uniform noise with infinity norm <= radius
            # torch.rand generates noise between 0 and 1
            # => This generates noise between -radius and radius
            noise = (
                torch.rand(batch_x.shape, device=device) * 2 * radius - radius
            )
            noisy_samples = batch_x + noise
            # Get new attributions from noisy samples
            noisy_attrs = _normalize_attrs(
                method(noisy_samples, batch_y).detach()
            )
            # Get relative norm of attribution difference
            # [batch_size]
            diffs.append(torch.norm(noisy_attrs.cpu() - attrs, dim=1).detach())
        # [batch_size, num_perturbations]
        diffs = torch.stack(diffs, 1)
        # [batch_size]
        result[method_name] = diffs.max(dim=1)[0].cpu()
    return result


def max_sensitivity(
    attributions_dataset: AttributionsDataset,
    batch_size: int,
    method_dict: Dict[str, AttributionMethod],
    num_perturbations: int,
    radius: float,
    device: torch.device = torch.device("cpu"),
) -> MaxSensitivityResult:
    """Computes the Max-Sensitivity metric for a given `Dataset` and attribution
    methods. Max-Sensitivity is computed by adding a small amount of uniform noise
    to the input samples and computing the norm of the difference in attributions
    between the original samples and the noisy samples.
    The maximum norm of difference is then taken as the Max-Sensitivity.

    The idea is that a small amount of noise should not change the attributions
    significantly, so the norm of the difference should be small. If the norm
    is large, then the attributions are not robust to small perturbations in the
    input.

    Parameters
    ----------
    attributions_dataset : Dataset
        The dataset for which the Max-Sensitivity should be computed.
    batch_size : int
        The batch size to use for computing the Max-Sensitivity.
    method_dict : Dict[str, AttributionMethod]
        Dictionary of attribution methods for which the Max-Sensitivity should be
        computed.
    num_perturbations : int
        The number of perturbations to use for computing the Max-Sensitivity.
    radius : float
        The radius of the uniform noise to add to the input samples.
    device : torch.device, optional
        Device to use, by default `torch.device("cpu")`.
    """
    grouped_dataset = GroupedAttributionsDataset(attributions_dataset)
    dataloader = DataLoader(
        grouped_dataset,
        batch_size=batch_size,
        num_workers=4,
        pin_memory=True,
    )

    method_names = list(method_dict.keys())
    result = MaxSensitivityResult(
        method_names,
        num_samples=len(grouped_dataset),
    )
    for batch_indices, batch_x, batch_y, batch_attr in dataloader:
        batch_x = batch_x.to(device)
        batch_y = batch_y.to(device)

        batch_result = _max_sensitivity_batch(
            batch_x,
            batch_y,
            batch_attr,
            method_dict,
            num_perturbations,
            radius,
            device,
        )
        result.add(GroupedBatchResult(batch_indices, batch_result))
    return result


import torch
from typing import Dict


class GroupedBatchResult:
    """
    Represents results from running a metric on a single batch of images.
    The results are grouped, i.e. the metric is computed for all attribution
    methods at a time. This is used for metrics  which have a shared computation
    for all attribution methods to save computation time, e.g. Infidelity.
    Used by GroupedMetricResult to handle results from distributed metrics.
    """
    def __init__(self, indices: torch.Tensor, results: Dict):
        self.results = results
        self.indices = indices



from typing import Optional, Tuple
import os
import torch
from torch.utils.data import Dataset

from attribench.distributed._distributed_computation import (
    DistributedComputation,
)
from attribench.distributed._message import PartialResultMessage
from attribench.distributed._worker import Worker, WorkerConfig
from attribench.functional._train_adversarial_patches import _make_patch
from attribench import ModelFactory


class PatchResult:
    def __init__(
        self, patch_label: int, val_loss: float, percent_successful: float
    ) -> None:
        self.patch_label = patch_label
        self.val_loss = val_loss
        self.percent_successful = percent_successful


class AdversarialPatchTrainingWorker(Worker):
    def __init__(
        self,
        worker_config: WorkerConfig,
        path: str,
        total_num_patches: int,
        batch_size: int,
        dataset: Dataset,
        model_factory: ModelFactory,
        labels: Optional[Tuple[int]] = None,
    ):
        super().__init__(worker_config)
        # Create a list of patch labels.
        # If the number of available labels is smaller than the number of
        # patches, the labels are repeated.
        if labels is None:
            labels = tuple(range(total_num_patches))
        num_repeats = total_num_patches // len(labels)
        labels = labels * (num_repeats + 1)

        # Each worker only trains a subset of the patches.
        rank = self.worker_config.rank
        world_size = self.worker_config.world_size
        self.patch_labels = labels[
            rank : total_num_patches : world_size
        ]
        self.dataset = dataset
        self.model_factory = model_factory
        self.batch_size = batch_size
        self.path = path

    def work(self):
        device = torch.device(self.worker_config.rank)
        model = self.model_factory()
        model.to(device)

        for patch_label in self.patch_labels:
            # Train patch
            patch, val_loss, percent_successful = _make_patch(
                self.dataset, self.batch_size, model, patch_label, device
            )

            # Save patch to disk
            torch.save(
                patch, os.path.join(self.path, f"patch_{patch_label}.pt")
            )

            # Send message to main process
            self.worker_config.send_result(
                PartialResultMessage(
                    self.worker_config.rank,
                    PatchResult(patch_label, val_loss, percent_successful),
                )
            )


class TrainAdversarialPatches(DistributedComputation):
    """Train adversarial patches for a given model and dataset and save
    them to disk. The patches are trained in parallel on multiple
    processes. Each process trains a subset of the patches.
    """

    def __init__(
        self,
        model_factory: ModelFactory,
        dataset: Dataset,
        num_patches: int,
        batch_size: int,
        path: str,
        labels: Optional[Tuple[int]] = None,
        address: str = "localhost",
        port: str = "12355",
        devices: Optional[Tuple[int]] = None,
    ):
        """
        Parameters
        ----------
        model_factory : ModelFactory
            ModelFactory instance or callable that returns a model.
            Used to create a model for each subprocess.
        dataset : Dataset
            Torch Dataset to use for training the patches.
        num_patches : int
            Number of patches to train.
        batch_size : int
            Batch size per subprocess to use for training.
        path : str
            Path to which the patches should be saved.
        labels : Optional[Tuple[int]], optional
            Tuple of labels to use for the patches.
            If `None`, the labels are assumed to be `range(num_patches)`.
            Default: `None`.
        address : str, optional
            Address for communication between subprocesses,
            by default "localhost"
        port : str, optional
            Port for communication between subprocesses, by default "12355"
        devices : Optional[Tuple[int]], optional
            Devices to use. If None, then all available devices are used.
            By default None.
        """
        super().__init__(address, port, devices)
        self.num_patches = num_patches
        self.labels = labels
        self.path = path
        self.prog = None
        self.model_factory = model_factory
        self.dataset = dataset
        self.batch_size = batch_size
        if not os.path.isdir(self.path):
            os.makedirs(self.path)

    def _create_worker(self, worker_config: WorkerConfig) -> Worker:
        return AdversarialPatchTrainingWorker(
            worker_config,
            self.path,
            self.num_patches,
            self.batch_size,
            self.dataset,
            self.model_factory,
            self.labels,
        )

    def _handle_result(self, result: PartialResultMessage[PatchResult]):
        # The workers save the files,
        # so no need to do anything except log results
        print(
            f"Received patch {result.data.patch_label}.",
            f"Loss: {result.data.val_loss:.3f}.",
            f"Success rate: {result.data.percent_successful:.3f}.",
        )



# Sources:
# https://towardsdatascience.com/better-heatmaps-and-correlation-matrix-plots-in-python-41445d0f2bec
# https://seaborn.pydata.org/generated/seaborn.heatmap.html
import seaborn as sns
from matplotlib import pyplot as plt
from matplotlib.figure import Figure
import numpy as np
import math


def heatmap(
    x,
    y,
    size,
    color,
    palette=None,
    figsize=(20, 20),
    glyph_scale=1500,
    fontsize=None,
    title=None,
    color_bounds=None,
    cbar=True,
    x_labels=None,
    y_labels=None,
) -> Figure:
    sns.set()
    fig = plt.figure(figsize=figsize)
    fig.suptitle(title, fontsize=16)
    plot_grid = plt.GridSpec(
        1, 15, hspace=0.2, wspace=0.1, figure=fig
    )  # 1x15 grid
    if cbar:
        ax = fig.add_subplot(
            plot_grid[:, :-1]
        )  # Use leftmost 14 columns for the main plot
    else:
        ax = fig.add_subplot()  # Use everything for the main plot

    # Mapping from colnames to integer coordinates
    if x_labels is None:
        x_labels = list(sorted(x.unique()))
    elif list(sorted(x_labels)) != list(sorted(x.unique())):
        raise ValueError("Invalid X labels")

    if y_labels is None:
        y_labels = list(sorted(y.unique()))[::-1]
    elif list(sorted(y_labels)) != list(sorted(y.unique())):
        raise ValueError("Invalid Y labels")

    x_to_num = {label: num for num, label in enumerate(x_labels)}
    y_to_num = {label: num for num, label in enumerate(y_labels)}

    if palette is None:
        palette = sns.diverging_palette(240, 10, n=256)
    color_min, color_max = (
        color_bounds
        if color_bounds is not None
        else (color.min(), color.max())
    )

    def value_to_color(val):
        val_position = float((val - color_min)) / (color_max - color_min)
        if math.isnan(
            val_position
        ):  # Might be nan if colors are all 0 (ie if nothing is significant)
            val_position = 0
        ind = int(val_position * (len(palette) - 1))
        return palette[ind]

    ax.scatter(
        x=x.map(x_to_num),  # Use mapping for x
        y=y.map(y_to_num),  # Use mapping for y
        s=(size * glyph_scale),  # Vector of square sizes
        c=color.apply(value_to_color),
        marker="s",  # Use square as scatterplot marker
    )

    # Force equal aspect such that all glyphs are square
    ax.set_aspect("equal", "box")

    # Show column labels on the axes
    ax.set_xticks([x_to_num[v] for v in x_labels])
    ax.set_xticklabels(
        x_labels, rotation=45, horizontalalignment="right", fontsize=fontsize
    )
    ax.set_yticks([y_to_num[v] for v in y_labels])
    ax.set_yticklabels(y_labels, fontsize=fontsize)
    ax.grid(False, "major")
    ax.grid(True, "minor")
    ax.set_xticks([t + 0.5 for t in ax.get_xticks()], minor=True)
    ax.set_yticks([t + 0.5 for t in ax.get_yticks()], minor=True)
    ax.set_xlim([-0.5, max(x_to_num.values()) + 0.5])
    ax.set_ylim([-0.5, max(y_to_num.values()) + 0.5])

    # Add color legend on the right side of the plot
    if cbar:
        legend_ax = fig.add_subplot(
            plot_grid[:, -1]
        )  # Use the rightmost column of the plot

        col_x = [0] * len(palette)  # Fixed x coordinate for the bars
        bar_y = np.linspace(
            color_min, color_max, len(palette)
        )  # y coordinates for each of the n_colors bars

        bar_height = bar_y[1] - bar_y[0]
        legend_ax.barh(
            y=bar_y,
            width=[5] * len(palette),  # Make bars 5 units wide
            left=col_x,  # Make bars start at 0
            height=bar_height,
            color=palette,
            linewidth=0,
        )
        legend_ax.set_xlim(1, 2)
        # Bars are going from 0 to 5,
        # so lets crop the plot somewhere in the middle
        legend_ax.grid(False)  # Hide grid
        legend_ax.set_facecolor("white")  # Make background white
        legend_ax.set_xticks([])  # Remove horizontal ticks
        legend_ax.set_yticks(
            np.linspace(min(bar_y), max(bar_y), 3)
        )  # Show vertical ticks for min, middle and max
        legend_ax.yaxis.tick_right()  # Show vertical ticks on the right
    return fig



from typing import Dict, Tuple
import pandas as pd
from attribench.plot._lib import effect_size_barplot
from attribench._stat import wilcoxon_tests


class WilcoxonBarPlot:
    """
    Alternative to WilcoxonSummaryPlot.
    Shows a bar plot of effect sizes, along with a grid showing statistical
    significance (green if significant, red otherwise).
    Provides more detailed information about effect size than the summary plot.
    Can get crowded if there are many methods/metrics.
    """

    def __init__(self, dfs: Dict[str, Tuple[pd.DataFrame, bool]]):
        self.dfs = dfs

    def render(self, figsize=(12, 6)):
        ALPHA = 0.01
        effect_sizes, pvalues = {}, {}

        for metric_name, (df, higher_is_better) in self.dfs.items():
            # Compute effect sizes and p-values
            mes, mpv = wilcoxon_tests(df, higher_is_better)
            effect_sizes[metric_name] = mes
            pvalues[metric_name] = mpv

        es_df = pd.DataFrame.from_dict(effect_sizes)
        pv_df = pd.DataFrame.from_dict(pvalues)
        fig, axs = effect_size_barplot(
            es_df, pv_df, self.dfs.keys(), ALPHA, figsize=figsize
        )
        return fig



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# Templates from: https://github.com/JamesALeedham/Sphinx-Autosummary-Recursion/tree/master

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
sys.path.insert(0, os.path.abspath('../..'))


# -- Project information -----------------------------------------------------

project = 'AttriBench'
copyright = '2023, Arne Gevaert, Axel-Jan Rousseau'
author = 'Arne Gevaert, Axel-Jan Rousseau'

# The full version, including alpha/beta/rc tags
release = '0.0.1'


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'numpydoc',  # Numpy documentation style
    'sphinx.ext.autodoc',  # Automatic documentation of docstrings
    'sphinx.ext.autosummary',  # Automatic generation of summary tables
    'sphinx.ext.viewcode',  # Add links to source code
    'sphinx_autodoc_typehints',  # Automatically document type hints for cleaner signatures
]
autoclass_content = 'both'
autodoc_inherit_docstrings = True
set_type_checking_flag = True
autodoc_default_options = {
    'members': True,
    'inherited-members': True,
}
numpydoc_show_class_members = False

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']
html_css_files = ['css/custom.css']


from ._infidelity import infidelity
from ._perturbation_generator import (
    PerturbationGenerator,
    NoisyBaselinePerturbationGenerator,
    GaussianPerturbationGenerator,
    SquarePerturbationGenerator,
    SegmentRemovalPerturbationGenerator,
)



import numpy as np
from typing import Dict, List

import torch
from attribench.data.attributions_dataset._attributions_dataset import (
    GroupedAttributionsDataset,
)
from attribench.masking import Masker
from .._metric_worker import GroupedMetricWorker, WorkerConfig
from attribench._model_factory import ModelFactory
from attribench.functional.metrics.sensitivity_n._sensitivity_n import (
    _sens_n_batch,
)


class SensitivityNWorker(GroupedMetricWorker):
    def __init__(
        self,
        worker_config: WorkerConfig,
        model_factory: ModelFactory,
        dataset: GroupedAttributionsDataset,
        batch_size: int,
        min_subset_size: float,
        max_subset_size: float,
        num_steps: int,
        num_subsets: int,
        maskers: Dict[str, Masker],
        activation_fns: List[str],
        segmented=False,
    ):
        super().__init__(worker_config, model_factory, dataset, batch_size)
        self.dataset = dataset
        self.activation_fns = activation_fns
        self.maskers: Dict[str, Masker] = maskers
        self.num_subsets = num_subsets
        self.num_steps = num_steps
        self.max_subset_size = max_subset_size
        self.min_subset_size = min_subset_size
        self.segmented = segmented
        n_range = np.linspace(
            self.min_subset_size, self.max_subset_size, self.num_steps
        )
        if segmented:
            n_range = n_range * 100
        else:
            total_num_features = np.prod(dataset.attributions_shape)
            n_range = n_range * total_num_features
        self.n_range = n_range.astype(int)

    def process_batch(
        self,
        batch_x: torch.Tensor,
        batch_y: torch.Tensor,
        batch_attr: Dict[str, torch.Tensor],
    ):
        return _sens_n_batch(
            batch_x,
            batch_y,
            self.model,
            batch_attr,
            self.maskers,
            self.activation_fns,
            self.n_range,
            self.num_subsets,
            self.segmented,
        )



from tqdm import tqdm
from ...data.attributions_dataset._attributions_dataset import (
    AttributionsDataset,
    GroupedAttributionsDataset,
)
from ... import MethodFactory, AttributionMethod, ModelFactory
from ...result import ParameterRandomizationResult
from ...result._grouped_batch_result import GroupedBatchResult
from typing import Dict, Callable
import torch
from torch.utils.data import DataLoader
from ..._stat import rowwise_spearmanr


def _randomize_parameters(model_factory: ModelFactory) -> torch.nn.Module:
    randomized_model = model_factory()
    for layer in randomized_model.modules():
        if hasattr(layer, "reset_parameters"):
            layer.reset_parameters()
    return randomized_model


def _parameter_randomization_batch(
    batch_x: torch.Tensor,
    batch_y: torch.Tensor,
    batch_attr: Dict[str, torch.Tensor],
    method_dict_rand: Dict[str, AttributionMethod],
    device: torch.device,
    agg_fn: Callable[
        [
            torch.Tensor,
            int,
        ],
        torch.Tensor,
    ]
    | None = None,
    agg_dim: int | None = None,
) -> Dict[str, torch.Tensor]:
    if set(method_dict_rand.keys()) != set(batch_attr.keys()):
        raise ValueError(
            "Method dictionary and batch attributions dictionary"
            " must have the same keys."
        )
    result: Dict[str, torch.Tensor] = {
        method_name: torch.zeros(1) for method_name in method_dict_rand.keys()
    }
    batch_x = batch_x.to(device)
    batch_y = batch_y.to(device)

    for method_name, method in method_dict_rand.items():
        # If attributions were aggregated, we need to perform the same
        # aggregations on the randomized attributions
        attrs_rand = method(batch_x, batch_y).detach().cpu()
        if agg_fn is not None:
            assert agg_dim is not None
            # agg_dim is expressed in terms of sample dimension, need to add
            # 1 to account for batch dimension
            attrs_rand = agg_fn(attrs_rand, agg_dim+1)
        attrs_rand = attrs_rand.flatten(1)

        attrs_orig = batch_attr[method_name].cpu().flatten(1)
        result[method_name] = torch.tensor(
            rowwise_spearmanr(attrs_rand.numpy(), attrs_orig.numpy())
        )

    return result


def parameter_randomization(
    model_factory: ModelFactory,
    attributions_dataset: AttributionsDataset,
    batch_size: int,
    method_factory: MethodFactory,
    device: torch.device = torch.device("cpu"),
) -> ParameterRandomizationResult:
    """
    Computes the Parameter Randomization metric for a given
    :class:`~attribench.data.AttributionsDataset`.

    The Parameter Randomization metric is computed by randomly re-initializing the
    parameters of the model and computing an attribution map of the prediction
    on the re-initialized model. The metric value is the spearman rank correlation
    between the original attribution map and the attribution map of the
    re-initialized model. If this value is high, then the attribution method is
    insensitive to the model parameters, thereby failing the sanity check.

    Source: Adebayo, J., Gilmer, J., Muelly, M., Goodfellow, I.,
    Hardt, M., & Kim, B. (2018). Sanity checks for saliency maps.
    Advances in neural information processing systems, 31.

    Parameters
    ----------
    model_factory : ModelFactory
        ModelFactory instance or callable that returns a model.
        Used to create the original model and a randomized copy.
    attributions_dataset : AttributionsDataset
        Dataset containing the samples and attributions to compute
        the Parameter Randomization metric for.
    batch_size : int
        Batch size to use when computing the metric.
    method_factory : MethodFactory
        MethodFactory instance or callable that returns a dictionary mapping
        method names to AttributionMethod objects.
    device : torch.device, optional
        Device to use when computing the metric, by default torch.device("cpu")

    Returns
    -------
    ParameterRandomizationResult
        Result of the Parameter Randomization metric computation.
    """
    
    randomized_model = _randomize_parameters(model_factory)
    randomized_model.to(device)
    randomized_model.eval()

    method_dict_rand = method_factory(randomized_model)
    grouped_dataset = GroupedAttributionsDataset(attributions_dataset)
    dataloader = DataLoader(
        grouped_dataset, batch_size=batch_size, num_workers=4, pin_memory=True
    )

    method_names = method_factory.get_method_names()
    result = ParameterRandomizationResult(
        method_names, num_samples=len(grouped_dataset)
    )

    for batch_indices, batch_x, batch_y, batch_attr in tqdm(dataloader):
        agg_fn = None
        agg_dim = None
        if attributions_dataset.aggregate_fn is not None:
            agg_fn = attributions_dataset.aggregate_fn
            agg_dim = attributions_dataset.aggregate_dim

        batch_result = _parameter_randomization_batch(
            batch_x,
            batch_y,
            batch_attr,
            method_dict_rand,
            device,
            agg_fn,
            agg_dim,
        )
        result.add(GroupedBatchResult(batch_indices, batch_result))
    return result



from typing import Tuple
from attribench._model_factory import ModelFactory
from attribench.data.attributions_dataset._attributions_dataset import AttributionsDataset, GroupedAttributionsDataset
from attribench.distributed.metrics._metric_worker import WorkerConfig
from .._metric import Metric
from ._parameter_randomization_worker import ParameterRandomizationWorker
from attribench.result import ParameterRandomizationResult
from attribench import MethodFactory


class ParameterRandomization(Metric):
    """
    Computes the Parameter Randomization metric for a given
    :class:`~attribench.data.AttributionsDataset` and model using multiple processes.

    The Parameter Randomization metric is computed by randomly re-initializing the
    parameters of the model and computing an attribution map of the prediction
    on the re-initialized model. The metric value is the spearman rank correlation
    between the original attribution map and the attribution map of the
    re-initialized model. If this value is high, then the attribution method is
    insensitive to the model parameters, thereby failing the sanity check.

    Source: Adebayo, J., Gilmer, J., Muelly, M., Goodfellow, I.,
    Hardt, M., & Kim, B. (2018). Sanity checks for saliency maps.
    Advances in neural information processing systems, 31.
    """

    def __init__(
        self,
        model_factory: ModelFactory,
        attributions_dataset: AttributionsDataset,
        batch_size: int,
        method_factory: MethodFactory,
        address="localhost",
        port=12355,
        devices: Tuple | None = None,
    ):
        """
        Parameters
        ----------
        model_factory : ModelFactory
            ModelFactory instance or callable that returns a model.
            Used to create a model for each subprocess, and to create a
            randomized copy of the model.
        attributions_dataset : AttributionsDataset
            Dataset containing the samples and attributions to compute
            the Parameter Randomization metric for.
        batch_size : int
            Batch size per subprocess to use when computing the metric.
        address : str
            Address to use for the distributed computation.
        port : str | int
            Port to use for the distributed computation.
        devices : Tuple | None, optional
            Tuple of devices to use for the distributed computation.
            If None, then all available devices are used.
        """
        super().__init__(
            model_factory,
            attributions_dataset,
            batch_size,
            address,
            port,
            devices,
        )
        self.dataset = GroupedAttributionsDataset(attributions_dataset)
        self._result = ParameterRandomizationResult(
            attributions_dataset.method_names, attributions_dataset.num_samples
        )
        self.method_factory = method_factory
        
        self.agg_fn = None
        self.agg_dim = None
        if attributions_dataset.aggregate_fn is not None:
            self.agg_fn = attributions_dataset.aggregate_fn
            self.agg_dim = attributions_dataset.aggregate_dim

    def _create_worker(
        self, worker_config: WorkerConfig
    ) -> ParameterRandomizationWorker:
        
        return ParameterRandomizationWorker(
            worker_config,
            self.model_factory,
            self.dataset,
            self.batch_size,
            self.method_factory,
            self.agg_fn,
            self.agg_dim,
        )



from torch.utils.data import Dataset
from .._typing import _check_is_dataset
from typing import Tuple
import h5py


class HDF5Dataset(Dataset):
    """
    Dataset stored in a HDF5 file.

    The HDF5 file must contain the following datasets:

    - ``samples: [num_samples, *sample_shape]``
    - ``labels: [num_samples]``
    """

    def __init__(self, path: str):
        """
        Parameters
        ----------
        path : str
            Path to the HDF5 file.
        """
        self.path = path
        self.file: h5py.File | None = None
        self._sample_shape: Tuple | None = None

    @property
    def sample_shape(self):
        if self.file is None:
            with h5py.File(self.path, "r") as fp:
                return _check_is_dataset(fp["samples"]).shape[1:]
        return _check_is_dataset(self.file["samples"]).shape[1:]

    def __getitem__(self, index):
        if self.file is None:
            self.file = h5py.File(self.path, "r")
        return (
            _check_is_dataset(self.file["samples"])[index],
            _check_is_dataset(self.file["labels"])[index],
        )

    def __len__(self):
        if self.file is None:
            with h5py.File(self.path, "r") as fp:
                return len(_check_is_dataset(fp["samples"]))
        return len(_check_is_dataset(self.file["samples"]))

    def __del__(self):
        if self.file is not None:
            self.file.close()



from attribench.masking import Masker
from attribench._segmentation import segment_attributions
from typing import List, Union, Optional
import numpy as np
import torch


class ImageMasker(Masker):
    def __init__(self, feature_level):
        if feature_level not in ("channel", "pixel"):
            raise ValueError(
                f"feature_level must be 'channel' or 'pixel'. Found {feature_level}."
            )
        self.feature_level = feature_level
        super().__init__()
        # will be set after initialize_batch:
        self.segmented_samples: Optional[torch.Tensor] = None
        self.segmented_attributions: Optional[np.ndarray] = None
        self.segment_indices: Optional[List[np.ndarray]] = None
        self.use_segments: bool = False

    def set_batch(
        self,
        samples: torch.Tensor,
        attributions: np.ndarray = None,
        segmented_samples: torch.Tensor = None,
    ):
        if attributions is not None and not self._check_attribution_shape(
            samples, attributions
        ):
            raise ValueError(
                f"samples and attribution shape not compatible for feature level {self.feature_level}."
                f"Found shapes {samples.shape} and {attributions.shape}"
            )
        if segmented_samples is not None:
            if not (
                samples.shape[0] == segmented_samples.shape[0]
                and samples.shape[-2:] == segmented_samples.shape[-2:]
            ):
                raise ValueError(
                    f"Incompatible shapes: {samples.shape}, {segmented_samples.shape}"
                )
            if samples.device != segmented_samples.device:
                raise ValueError(
                    "Device for samples and segmented_samples must be equal."
                    f"Got {samples.device} for samples, {segmented_samples.device} for segmented_samples."
                )

        self.samples = samples
        self.attributions = attributions
        self.segmented_samples = segmented_samples
        if segmented_samples is not None:
            self.segment_indices = [
                np.unique(segmented_samples.cpu().numpy()[i, ...])
                for i in range(samples.shape[0])
            ]
        self.use_segments = self.segmented_samples is not None

        if (
            self.segmented_samples is not None
            and self.attributions is not None
        ):
            self.segmented_attributions = segment_attributions(
                self.segmented_samples.cpu().numpy(), self.attributions
            )
            sorted_indices = self.segmented_attributions.argsort()

            # Filter out the -np.inf values from the sorted indices
            filtered_sorted_indices = []
            for i in range(self.segmented_samples.shape[0]):
                num_infs = np.count_nonzero(
                    self.segmented_attributions[i, ...] == -np.inf
                )
                filtered_sorted_indices.append(sorted_indices[i, num_infs:])
            self.sorted_indices = filtered_sorted_indices

        elif self.attributions is not None:
            self.sorted_indices = self.attributions.reshape(
                self.attributions.shape[0], -1
            ).argsort()

        self.initialize_baselines(self.samples)

    def get_num_features(self):
        if self.use_segments:
            raise ValueError(
                "When using segments, total number of features varies per image."
            )
        if self.feature_level == "channel":
            return self.samples.flatten(1).shape[-1]
        if self.feature_level == "pixel":
            return self.samples.flatten(2).shape[-1]

    def mask_top(self, k):
        if not self.use_segments:
            return super().mask_top(k)
        if k == 0:
            return self.samples
        # When using segments, k is relative (between 0 and 1)
        indices = []
        for i in range(self.samples.shape[0]):
            num_segments = len(self.sorted_indices[i])
            num_to_mask = int(num_segments * k)
            indices.append(self.sorted_indices[i][-num_to_mask:])
        return self._mask_segments(indices)

    def mask_bot(self, k):
        if not self.use_segments:
            return super().mask_bot(k)
        if k == 0:
            return self.samples
        # When using segments, k is relative (between 0 and 1)
        indices = []
        for i in range(self.samples.shape[0]):
            num_segments = len(self.sorted_indices[i])
            num_to_mask = int(num_segments * k)
            indices.append(self.sorted_indices[i][:num_to_mask])
        return self._mask_segments(indices)

    def mask_rand(self, k, return_indices=False):
        if k == 0:
            return self.samples
        rng = np.random.default_rng()
        if not self.use_segments:
            return super().mask_rand(k, return_indices)
        else:
            # this is done a little different form super(),
            # no shuffle here: for each image, only select segments that exist in that image
            # k can be large -> rng.choice raises exception if k > number of segments
            indices = [
                rng.choice(self.segment_indices[i], size=k, replace=False)
                for i in range(self.segmented_samples.shape[0])
            ]
            masked_samples = self._mask_segments(indices)
        if return_indices:
            return masked_samples, indices
        return masked_samples

    def _check_attribution_shape(
        self, samples: torch.Tensor, attributions: np.ndarray
    ):
        if self.feature_level == "channel":
            # Attributions should be same shape as samples
            return list(samples.shape) == list(attributions.shape)
        elif self.feature_level == "pixel":
            # attributions should have the same shape as samples,
            # except the channel dimension must be 1
            aggregated_shape = list(samples.shape)
            aggregated_shape[1] = 1
            return aggregated_shape == list(attributions.shape)

    def _mask(self, indices: np.ndarray):
        if self.baseline is None:
            raise ValueError("Masker was not initialized.")
        if self.use_segments:
            return self._mask_segments(indices)
        else:
            batch_size, num_channels, rows, cols = self.samples.shape
            num_indices = indices.shape[1]
            batch_dim = np.tile(
                range(batch_size), (num_indices, 1)
            ).transpose()

            to_mask = torch.zeros(
                self.samples.shape, device=self.samples.device
            ).flatten(1 if self.feature_level == "channel" else 2)
            # to_mask = np.zeros(self.samples.shape)
            if self.feature_level == "channel":
                to_mask[batch_dim, indices] = 1.0
            else:
                try:
                    to_mask[batch_dim, :, indices] = 1.0
                except IndexError:
                    raise ValueError(
                        "Masking index was out of bounds. "
                        "Make sure the masking policy is compatible with method output."
                    )
            return self._mask_boolean(to_mask.view(self.samples.shape).bool())

    def _mask_segments(
        self, segments: Union[np.ndarray, List[np.ndarray]]
    ) -> torch.tensor:
        if not self.segmented_samples.shape[0] == len(segments):
            raise ValueError(
                f"Number of segment lists doesn't match number of images: {self.segmented_samples.shape[0]}"
                f"were expected, {len(segments)} were given."
            )
        bool_masks = []
        for i in range(self.samples.shape[0]):
            seg_img = self.segmented_samples[i, ...]
            segs = segments[i]
            bool_masks.append(
                _isin(seg_img, torch.tensor(segs, device=seg_img.device))
            )
        bool_masks = torch.stack(bool_masks, dim=0)
        if self.samples.shape[1] == 3:
            bool_masks = bool_masks.repeat(1, 3, 1, 1)
        return self._mask_boolean(bool_masks)

    def _mask_boolean(self, bool_mask: torch.tensor):
        return (
            self.samples
            - (bool_mask * self.samples)
            + (bool_mask * self.baseline)
        )

    def initialize_baselines(self, samples: torch.Tensor):
        raise NotImplementedError


def _isin(a: torch.Tensor, b: torch.Tensor):
    # https://stackoverflow.com/questions/60918304/get-indices-of-elements-in-tensor-a-that-are-present-in-tensor-b
    return (a[..., None] == b).any(-1)



import torch
from typing import Dict, List


class BatchResult:
    """
    Represents results from running a metric on a single batch of images.
    Used by MetricResult to handle results from distributed metrics.
    """
    def __init__(self, indices: torch.Tensor, results: Dict,
                 method_names: List[str]):
        self.method_names = method_names
        self.results = results
        self.indices = indices



from ._message import PartialResultMessage
from ._distributed_computation import DistributedComputation
from ._distributed_sampler import DistributedSampler
from ._worker import Worker, WorkerConfig
from attribench._model_factory import ModelFactory

from attribench._method_factory import MethodFactory
from attribench.data import AttributionsDatasetWriter, IndexDataset
from torch.utils.data import Dataset, DataLoader
from typing import Tuple, Optional
import torch
from numpy import typing as npt
from tqdm import tqdm


class AttributionResult:
    def __init__(
        self, indices: npt.NDArray, attributions: npt.NDArray, method_name: str
    ):
        self.indices = indices
        self.attributions = attributions
        self.method_name = method_name


class AttributionsWorker(Worker):
    def __init__(
        self,
        worker_config: WorkerConfig,
        model_factory: ModelFactory,
        method_factory: MethodFactory,
        dataset: IndexDataset,
        batch_size: int,
    ):
        super().__init__(worker_config)
        self.batch_size = batch_size
        self.dataset = dataset
        self.method_factory = method_factory
        self.model_factory = model_factory

    def work(self):
        sampler = DistributedSampler(
            self.dataset,
            self.worker_config.world_size,
            self.worker_config.rank,
            shuffle=False,
        )
        dataloader = DataLoader(
            self.dataset,
            sampler=sampler,
            batch_size=self.batch_size,
            num_workers=4,
            pin_memory=True,
        )
        device = torch.device(self.worker_config.rank)
        model = self.model_factory()
        model.to(device)
        method_dict = self.method_factory(model)

        for batch_indices, batch_x, batch_y in dataloader:
            batch_x = batch_x.to(device)
            batch_y = batch_y.to(device)
            for method_name, method in method_dict.items():
                with torch.no_grad():
                    attrs = method(batch_x, batch_y)
                    result = AttributionResult(
                        batch_indices.cpu().numpy(),
                        attrs.cpu().numpy(),
                        method_name,
                    )
                    self.worker_config.send_result(
                        PartialResultMessage(self.worker_config.rank, result)
                    )


class ComputeAttributions(DistributedComputation):
    """Compute attributions for a dataset using multiple processes.
    The attributions are written to a HDF5 file.
    The number of processes is determined by the number of devices.
    If no devices are specified, then all available devices are used.
    Samples are distributed evenly across the processes.

    If you want to compute attributions and simply return them, rather than
    storing them in a file, then use the
    :func:`~attribench.functional.compute_attributions` function instead.
    """

    def __init__(
        self,
        model_factory: ModelFactory,
        method_factory: MethodFactory,
        dataset: Dataset,
        batch_size: int,
        address="localhost",
        port="12355",
        devices: Optional[Tuple] = None,
    ):
        """
        Parameters
        ----------
        model_factory : ModelFactory
            ModelFactory instance or callable that returns a model.
            Used to create a model for each subprocess.
        method_factory : MethodFactory
            MethodFactory instance or callable that returns a dictionary of
            attribution methods, given a model.
        dataset : Dataset
            Torch Dataset to use for computing the attributions.
        batch_size : int
            The batch size to use for computing the attributions.
        writer : AttributionsDatasetWriter
            AttributionsDatasetWriter to write the attributions to.
        address : str, optional
            Address to use for the multiprocessing connection.
            By default "localhost".
        port : str, optional
            Port to use for the multiprocessing connection.
            By default "12355".
        devices : Optional[Tuple], optional
            Devices to use. If None, then all available devices are used.
            By default None.
        """
        super().__init__(address, port, devices)
        self.model_factory = model_factory
        self.method_factory = method_factory
        self.dataset = IndexDataset(dataset)
        self.batch_size = batch_size
        self.prog: tqdm | None = None
        self.writer: AttributionsDatasetWriter | None = None

    def run(self, path: str):
        """Run the computation.

        Parameters
        ----------
        path : str
            Path to the HDF5 file to write the attributions to.
        """
        self.writer = AttributionsDatasetWriter(
            path,
            num_samples=len(self.dataset),
        )
        self.prog = tqdm(total=len(self.dataset) * len(self.method_factory))
        super().run()
    
    def _cleanup(self):
        if self.prog is not None:
            self.prog.close()

    def _create_worker(
        self, worker_config: WorkerConfig
    ) -> Worker:
        return AttributionsWorker(
            worker_config,
            self.model_factory,
            self.method_factory,
            self.dataset,
            self.batch_size,
        )

    def _handle_result(
        self, result_message: PartialResultMessage[AttributionResult]
    ):
        assert self.writer is not None
        indices = result_message.data.indices
        attributions = result_message.data.attributions
        method_name = result_message.data.method_name
        self.writer.write(indices, attributions, method_name)
        if self.prog is not None:
            self.prog.update(len(indices))



from torch.utils.data import Dataset, Sampler
from torch.utils.data.sampler import T_co
from typing import Iterator
import random


class DistributedSampler(Sampler):
    def __init__(self, dataset: Dataset, world_size: int, rank: int, shuffle=True):
        """
        DistributedSampler is an alternative to the PyTorch DistributedSampler that
        does not add extra samples to make the total count evenly divisible.
        Credit: https://github.com/SeungjunNah/DeepDeblur-PyTorch/blob/master/src/data/sampler.py
        """
        super().__init__(None)
        self.dataset = dataset
        self.world_size = world_size
        self.rank = rank

        ds_size = len(dataset)
        self.indices = list(range(ds_size))[self.rank:ds_size:self.world_size]
        if shuffle:
            random.shuffle(self.indices)

    def __iter__(self) -> Iterator[T_co]:
        return iter(self.indices)



import torch
from attribench.masking import Masker
from abc import abstractmethod


class MinimalSubsetDataset:
    def __init__(
        self,
        num_steps,
        samples: torch.Tensor,
        attrs: torch.Tensor,
        masker: Masker,
    ):
        self.num_steps = num_steps
        self.samples = samples
        self.masker = masker
        masker.set_batch(samples, attrs)
        self.total_features = self.masker.get_num_features()
        self.step_size = int(self.total_features / num_steps)
        if num_steps > self.total_features or num_steps < 2:
            raise ValueError(
                f"Number of steps must be between 2 and {self.total_features}"
                f" (got {num_steps})"
            )

    def __len__(self):
        return self.num_steps

    @abstractmethod
    def __getitem__(self, item):
        raise NotImplementedError


class MinimalSubsetDeletionDataset(MinimalSubsetDataset):
    def __getitem__(self, item):
        # Mask the k most important pixels
        num_to_mask = self.step_size * item
        if num_to_mask > self.total_features:
            raise StopIteration
        masked_samples = self.masker.mask_top(num_to_mask)
        return masked_samples, num_to_mask


class MinimalSubsetInsertionDataset(MinimalSubsetDataset):
    def __getitem__(self, item):
        # Mask the n-k least important pixels
        num_to_insert = self.step_size * item
        num_to_mask = self.total_features - num_to_insert
        if num_to_mask > self.total_features:
            raise StopIteration
        masked_samples = self.masker.mask_bot(num_to_mask)
        return masked_samples, num_to_insert



from attribench.masking.image import ImageMasker
from cv2 import blur
import numpy as np
import torch


class BlurringImageMasker(ImageMasker):
    """Image masker that masks pixels or features by replacing them with
    a blurred version of the image. The amount of blurring is controlled
    by the ``kernel_size`` parameter, which is expressed as a fraction
    of the image height.
    """
    def __init__(self, masking_level: str, kernel_size: float):
        """
        Parameters
        ----------
        feature_level : str
            The level at which to mask the image. Must be either
            ``"pixel"`` or ``"feature"``.
        kernel_size : float
            Kernel size for the blurring operation, expressed as a fraction
            of the image height. Must be between 0 and 1.

        Raises
        ------
        ValueError
            If ``kernel_size`` is not between 0 and 1.
        """
        super().__init__(masking_level)
        if not 0 < kernel_size < 1.0:
            raise ValueError(
                "Kernel size is expressed as a fraction of image height,"
                " and must be between 0 and 1."
            )
        self.kernel_size = kernel_size

    def _initialize_baselines(self, samples: torch.Tensor):
        kernel_size = int(self.kernel_size * samples.shape[-1])

        baseline = []
        for i in range(samples.shape[0]):
            sample = samples[i, ...].cpu().numpy()
            cv_sample = np.transpose(sample, (1, 2, 0))
            blurred_sample = blur(cv_sample, (kernel_size, kernel_size))
            if len(blurred_sample.shape) == 2:
                blurred_sample = blurred_sample[..., np.newaxis]
            baseline.append(np.transpose(blurred_sample, (2, 0, 1)))
        self.baseline = torch.tensor(
            np.stack(baseline, axis=0), device=samples.device
        )



from typing import Dict, Tuple
import pandas as pd
from attribench.plot._lib import heatmap
from attribench._stat import wilcoxon_tests
import seaborn as sns


class WilcoxonSummaryPlot:
    def __init__(self, dfs: Dict[str, Tuple[pd.DataFrame, bool]]):
        self.dfs = dfs

    def render(
        self,
        figsize=(20, 20),
        glyph_scale=1500,
        fontsize=None,
        title=None,
        method_order=None,
    ):
        pvalues, effect_sizes = {}, {}
        for metric_name, (df, higher_is_better) in self.dfs.items():
            mes, mpv = wilcoxon_tests(df, higher_is_better)
            effect_sizes[metric_name] = mes
            pvalues[metric_name] = mpv
        pvalues = pd.DataFrame(pvalues)
        effect_sizes = pd.DataFrame(effect_sizes).abs()
        effect_sizes[pvalues > 0.01] = 0

        # Normalize each column of the effect sizes dataframe
        effect_sizes = effect_sizes / effect_sizes.max()
        effect_sizes = effect_sizes.fillna(0)

        effect_sizes = pd.melt(effect_sizes.reset_index(), id_vars="index")
        effect_sizes.columns = ["method", "metric", "value"]

        return heatmap(
            x=effect_sizes["method"],
            y=effect_sizes["metric"],
            size=effect_sizes["value"],
            color=effect_sizes["value"],
            palette=sns.color_palette("Greens", n_colors=256),
            figsize=figsize,
            glyph_scale=glyph_scale,
            fontsize=fontsize,
            title=title,
            cbar=False,
            x_labels=method_order,
        )



import pandas as pd
import matplotlib.pyplot as plt
from matplotlib.figure import Figure
from attribench.plot import Plot


def mad_ratio(df):
    # Basically the F-ratio but using MAD instead of SS
    # MAD = median of absolute deviation to median
    # Mean of MADs for each method
    group_medians = df.median()
    within_mad = df.sub(group_medians).abs().median().mean()

    # MAD of group medians to global median
    global_median = df.stack().median()
    # between_mad = group_medians.sub(global_median).abs().median()
    between_mad = df.stack().sub(global_median).abs().median()

    return between_mad / within_mad


class MADRatioPlot(Plot):
    """
    Bar plot of the MAD (Median Absolute Deviation) ratio for each metric.
    The MAD ratio is the ratio of the total MAD to the MAD within groups.
    If this ratio is high, it means that there are differences in method
    behaviour.
    If the ratio is close to 1, it means that the methods behave similarly.
    This can be viewed as a one-way ANOVA test, but using non-parametric MAD
    instead of sum-of-squares.
    """

    def render(
        self, title: str | None = None, figsize=(10, 10), fontsize=20
    ) -> Figure:
        """Render the plot.

        Parameters
        ----------
        title : str | None, optional
            Title of the figure, by default None
        figsize : tuple, optional
            Size of the figure, by default (10, 10)
        fontsize : int, optional
            Font size of x and y axis ticks, by default 20

        Returns
        -------
        Figure
            Rendered Matplotlib figure.
        """
        ratios = {}
        for metric_name, (df, inverted) in self.dfs.items():
            ratios[metric_name] = mad_ratio(df)
        df = pd.DataFrame(ratios, index=["MAD Ratio"]).transpose()
        fig, ax = plt.subplots()
        plt.xticks(fontsize=fontsize)
        plt.yticks(fontsize=fontsize)
        df.plot.barh(figsize=figsize, ax=ax)
        if title is not None:
            ax.set_title(title)
        fig.tight_layout()
        return fig



import random
import logging
import re
import os
from typing import Dict
from torch import nn
from torch.utils.data import Dataset, DataLoader
from attribench._attribution_method import AttributionMethod
from attribench.result import ImpactCoverageResult
from attribench.result._grouped_batch_result import GroupedBatchResult
from attribench.data import IndexDataset
import torch
from itertools import cycle
import numpy as np


def _impact_coverage_batch(
    model: nn.Module,
    method_dict: Dict[str, AttributionMethod],
    batch_x: torch.Tensor,
    batch_y: torch.Tensor,
    patch_folder: str,
    patch_names_cycle: cycle,
    device: torch.device,
) -> Dict[str, torch.Tensor]:
    target_expr = re.compile(r".*_([0-9]*)\.pt")
    batch_result: Dict[str, torch.Tensor] = {
        method_name: torch.zeros(1) for method_name in method_dict.keys()
    }
    batch_x = batch_x.to(device)
    batch_y = batch_y.to(device)

    # Get original output and initialize datastructures
    with torch.no_grad():
        original_output = model(batch_x).detach().cpu()
    successful = torch.zeros(batch_x.shape[0]).bool()
    attacked_samples = batch_x.clone()
    targets = torch.zeros(batch_y.shape).long()
    patch_mask = torch.zeros(batch_x.shape)
    max_tries = 50
    num_tries = 0

    # Apply patches to images
    while not torch.all(successful):
        num_tries += 1
        # Load next patch
        patch_name = next(patch_names_cycle)
        match_expr = target_expr.match(patch_name)
        if match_expr is None:
            raise ValueError(
                f"Patch name {patch_name} does not match" " expected format."
            )
        target = int(match_expr.group(1))
        patch = torch.load(
            os.path.join(patch_folder, patch_name),
            map_location=lambda storage, _: storage,
        ).to(device)
        image_size = batch_x.shape[-1]
        patch_size = patch.shape[-1]

        # Apply patch to all images in batch (random location,
        # but same for each image in batch)
        indx = random.randint(0, image_size - patch_size)
        indy = random.randint(0, image_size - patch_size)
        attacked_samples[~successful, ...] = batch_x[~successful, ...].clone()
        attacked_samples[
            ~successful,
            :,
            indx : indx + patch_size,
            indy : indy + patch_size,
        ] = patch.float()
        with torch.no_grad():
            adv_out = model(attacked_samples).detach().cpu()

        # Set the patch mask and targets for the samples that were
        # successful this iteration
        # We set the patch mask for all samples that weren't yet
        # successful
        # This way, if any samples can't be attacked,
        # they will still have a patch on them
        # (even though it didn't flip the prediction)
        patch_mask[~successful, ...] = 0
        patch_mask[
            ~successful,
            :,
            indx : indx + patch_size,
            indy : indy + patch_size,
        ] = 1
        targets[~successful] = target

        # Add the currently successful samples to all successful samples
        successful_now = (
            # Output was originally not equal to target
            (original_output.argmax(axis=1) != target)
            # Output is now equal to target
            & (adv_out.argmax(axis=1) == target)
            # Ground truth is not equal to target
            & (batch_y.cpu() != target)
        )
        successful = successful | successful_now

        if num_tries > max_tries:
            logging.warning(
                "Not all samples could be attacked:"
                f"{torch.sum(successful)}/{batch_x.size(0)}"
                " were successful."
            )
            break
    targets = targets.to(device)

    # Compute impact coverage for each method
    for method_name, method in method_dict.items():
        attrs = method(attacked_samples, targets).detach().cpu().numpy()

        # Check attributions shape
        if attrs.shape[1] not in (1, 3):
            raise ValueError(
                "Impact Coverage only works on image data."
                "Attributions must have 1 or 3 color channels."
                f"Found attributions shape {attrs.shape}."
            )
        # If attributions have only 1 color channel,
        # we need a single-channel patch mask as well
        if attrs.shape[1] == 1:
            patch_mask = patch_mask[:, 0, :, :]

        # Get indices of top k attributions
        flattened_attrs = attrs.reshape(attrs.shape[0], -1)
        sorted_indices = flattened_attrs.argsort()
        # Number of top attributions is equal to number of features
        # masked by the patch
        # We assume here that the mask is the same size for all samples!
        nr_top_attributions = patch_mask[0, ...].long().sum().item()

        # Create mask of critical factors (most important
        # pixels/features according to attributions)
        to_mask = sorted_indices[:, -nr_top_attributions:]
        critical_factor_mask = np.zeros(attrs.shape).reshape(
            attrs.shape[0], -1
        )
        batch_size = attrs.shape[0]
        batch_dim = np.tile(
            range(batch_size), (nr_top_attributions, 1)
        ).transpose()
        critical_factor_mask[batch_dim, to_mask] = 1
        critical_factor_mask = critical_factor_mask.astype(bool)

        # Calculate IoU of critical factors (top n attributions) with
        # adversarial patch
        patch_mask_flattened = patch_mask.flatten(1).bool().numpy()
        intersection = (patch_mask_flattened & critical_factor_mask).sum(
            axis=1
        )
        union = (patch_mask_flattened | critical_factor_mask).sum(axis=1)
        iou = intersection.astype(float) / union.astype(float)
        batch_result[method_name] = iou
    return batch_result


def impact_coverage(
    model: nn.Module,
    samples_dataset: Dataset,
    batch_size: int,
    method_dict: Dict[str, AttributionMethod],
    patch_folder: str,
    device: torch.device = torch.device("cpu"),
) -> ImpactCoverageResult:
    """Computes the Impact Coverage metric for a given dataset, model, and
    set of attribution methods.

    Impact Coverage is computed by applying an adversarial patch to the input.
    This patch causes the model to change its prediction.
    The Impact Coverage metric is the intersection over union (IoU) of the
    patch with the top n attributions of the input, where n is the number of
    features masked by the patch. The idea is that, as the patch causes the
    model to change its prediction, the corresponding region in the image
    should be highly relevant to the model's prediction.

    Impact Coverage requires a folder containing adversarial patches. The
    patches should be named as follows: patch_<target>.pt, where <target>
    is the target class of the patch. The target class is the class that
    the model will predict when the patch is applied to the input.

    To generate adversarial patches, the 
    :meth:`~attribench.functional.train_adversarial_patches` function
    or :class:`~attribench.distributed.TrainAdversarialPatches` class
    can be used.

    Parameters
    ----------
    model : nn.Module
        Model to compute Impact Coverage for.
    samples_dataset : Dataset
        Dataset to compute Impact Coverage for.
    batch_size : int
        Batch size to use when computing Impact Coverage.
    method_dict : Dict[str, AttributionMethod]
        Dictionary mapping method names to attribution methods.
    patch_folder : str
        Path to folder containing adversarial patches.
    device : torch.device, optional
        Device to use for computing Impact Coverage.
        Default: torch.device("cpu")

    Returns
    -------
    ImpactCoverageResult
        Result of the Impact Coverage metric.
    """
    # Get names of patches and compile regular expression for deriving
    # target labels
    patch_names = [
        filename
        for filename in os.listdir(patch_folder)
        if filename.endswith(".pt")
    ]
    patch_names_cycle = cycle(patch_names)

    index_dataset = IndexDataset(samples_dataset)
    dataloader = DataLoader(
        index_dataset, batch_size=batch_size, num_workers=4, pin_memory=True
    )

    result = ImpactCoverageResult(list(method_dict.keys()), len(index_dataset))

    for batch_indices, batch_x, batch_y in dataloader:
        batch_result = _impact_coverage_batch(
            model,
            method_dict,
            batch_x,
            batch_y,
            patch_folder,
            patch_names_cycle,
            device,
        )
        result.add(GroupedBatchResult(batch_indices, batch_result))
    return result



from ._distributed_sampler import DistributedSampler
from ._distributed_computation import DistributedComputation
from ._message import PartialResultMessage
from ._worker import Worker, WorkerConfig
from attribench import ModelFactory
from tqdm import tqdm
from attribench.data import HDF5DatasetWriter
from attribench.functional._select_samples import _select_samples_batch
from typing import Callable, Tuple, Optional
from torch.utils.data import Dataset, DataLoader
import torch
from numpy import typing as npt
from torch import nn
from multiprocessing.synchronize import Event


class SamplesResult:
    def __init__(self, samples: npt.NDArray, labels: npt.NDArray):
        self.samples = samples
        self.labels = labels


class SampleSelectionWorker(Worker):
    def __init__(
        self,
        worker_config: WorkerConfig,
        sufficient_samples: Event,
        batch_size: int,
        dataset: Dataset,
        model_factory: Callable[[], nn.Module],
    ):
        super().__init__(worker_config)
        self.sufficient_samples = sufficient_samples
        self.model_factory = model_factory
        self.dataset = dataset
        self.batch_size = batch_size

    def work(self):
        sampler = DistributedSampler(
            self.dataset,
            self.worker_config.world_size,
            self.worker_config.rank,
        )
        dataloader = DataLoader(
            self.dataset,
            sampler=sampler,
            batch_size=self.batch_size,
            num_workers=4,
            pin_memory=True,
        )
        device = torch.device(self.worker_config.rank)
        model = self.model_factory()
        model.to(device)
        it = iter(dataloader)

        for batch_x, batch_y in it:
            correct_samples, correct_labels = _select_samples_batch(
                batch_x, batch_y, model, device
            )
            result = SamplesResult(
                correct_samples.cpu().numpy(), correct_labels.cpu().numpy()
            )
            self.worker_config.send_result(
                PartialResultMessage(self.worker_config.rank, result)
            )
            if self.sufficient_samples.is_set():
                break


class SelectSamples(DistributedComputation):
    """Select correctly classified samples from a dataset and write them
    to a HDF5 file. This is done in a distributed fashion, i.e. each
    subprocess selects a part of the samples and writes them to the
    HDF5 file. The number of processes is determined by the number of
    devices.

    If you want to select correctly classified samples and return them,
    rather than storing them to a HDF5 file, use
    :func:`attribench.functional.select_samples` instead.
    """

    def __init__(
        self,
        model_factory: ModelFactory,
        dataset: Dataset,
        num_samples: int,
        batch_size: int,
        address: str = "localhost",
        port: str = "12355",
        devices: Optional[Tuple] = None,
    ):
        """
        Parameters
        ----------
        model_factory : ModelFactory
            ModelFactory instance or callable that returns a model.
            Used to instantiate a model for each subprocess.
        dataset : Dataset
            Torch Dataset containing the samples and labels.
        writer : HDF5DatasetWriter
            Writer to write the samples and labels to.
        num_samples : int
            Number of correctly classified samples to select.
        batch_size : int
            Batch size per subprocess to use for the dataloader.
        address : str, optional
            Address to use for the multiprocessing connection,
            by default "localhost"
        port : str, optional
            Port to use for the multiprocessing connection, by default "12355"
        devices : Tuple, optional
            Devices to use. If None, then all available devices are used.
            By default None.
        """
        super().__init__(address, port, devices)
        self.model_factory = model_factory
        self.dataset = dataset
        self.num_samples = num_samples
        self.batch_size = batch_size
        self.sufficient_samples = self.ctx.Event()
        self.count = 0
        self.prog: tqdm | None = None
        self.writer: HDF5DatasetWriter | None = None

    def _create_worker(
        self, worker_config: WorkerConfig
    ):
        return SampleSelectionWorker(
            worker_config,
            self.sufficient_samples,
            self.batch_size,
            self.dataset,
            self.model_factory,
        )

    def run(self, path: str):
        """Run the sample selection.

        Parameters
        ----------
        path : str
            Path to the HDF5 file to write the samples to.
        """
        self.writer = HDF5DatasetWriter(path, self.num_samples)
        self.prog = tqdm(total=self.num_samples)
        super().run()

    def _handle_result(self, result: PartialResultMessage[SamplesResult]):
        assert self.writer is not None

        # If too many samples, truncate
        samples = result.data.samples
        labels = result.data.labels
        if self.count + samples.shape[0] > self.num_samples:
            samples = samples[: self.num_samples - self.count]
            labels = labels[: self.num_samples - self.count]
            self.sufficient_samples.set()

        # Write to disk
        self.writer.write(samples, labels)

        # Update progress bar
        self.count += samples.shape[0]
        if self.prog is not None:
            self.prog.update(samples.shape[0])



from collections.abc import Sized
from torch.utils.data import Dataset


class IndexDataset(Dataset):
    """Wraps a dataset to return the index of the sample as well.
    Used internally to keep track of the index of a sample in a dataset.
    """

    def __init__(self, dataset: Dataset):
        """
        Parameters
        ----------
        dataset : Dataset
            The PyTorch Dataset to wrap.
        """
        super().__init__()
        self.dataset: Dataset = dataset

    def __len__(self):
        if isinstance(self.dataset, Sized):
            return len(self.dataset)
        else:
            raise ValueError("Base dataset has no __len__")

    def __getitem__(self, item):
        data, target = self.dataset[item]
        return item, data, target



import torch
from attribench.data.attributions_dataset._attributions_dataset import (
    GroupedAttributionsDataset,
)
from .._metric_worker import GroupedMetricWorker, WorkerConfig
from typing import Callable, Dict
from torch import nn
from attribench.functional.metrics._max_sensitivity import (
    _max_sensitivity_batch,
)
from attribench._method_factory import MethodFactory
from attribench._model_factory import ModelFactory


class MaxSensitivityWorker(GroupedMetricWorker):
    def __init__(
        self,
        worker_config: WorkerConfig,
        model_factory: ModelFactory,
        dataset: GroupedAttributionsDataset,
        batch_size: int,
        method_factory: MethodFactory,
        num_perturbations: int,
        radius: float,
    ):
        super().__init__(
            worker_config,
            model_factory,
            dataset,
            batch_size,
        )
        self.method_factory = method_factory
        self.num_perturbations = num_perturbations
        self.radius = radius

    def setup(self):
        self.model = self._get_model()
        self.method_dict = self.method_factory(self.model)

    def process_batch(
        self,
        batch_x: torch.Tensor,
        batch_y: torch.Tensor,
        batch_attr: Dict[str, torch.Tensor],
    ):
        return _max_sensitivity_batch(
            batch_x,
            batch_y,
            batch_attr,
            self.method_dict,
            self.num_perturbations,
            self.radius,
            self.device,
        )



from abc import abstractmethod
from typing import Tuple, Dict, List
import h5py
import numpy as np
from attribench.data.nd_array_tree._random_access_nd_array_tree import (
    RandomAccessNDArrayTree,
)
from ._batch_result import BatchResult
from attribench import result
import pandas as pd
import os
import yaml


class MetricResult:
    """Abstract class to represent results of distributed metrics."""

    def __init__(
        self,
        method_names: List[str],
        shape: List[int],
        levels: Dict[str, List[str]],
        level_order: List[str],
    ):
        """
        Parameters
        ----------
        method_names : Tuple[str, ...] | List[str]
            Names of attribution methods tested by the metric.
        shape : Tuple[int, ...] | List[int]
            Shape of numpy arrays that contain the results.
            Note that this is the result on the full dataset, not a single
            sample. For example, if the metric is computed on 100 samples and
            the metric returns 10 values per sample, then the shape
            should be ``(100, 10)``.
        levels : Dict[str, Tuple[str, ...] | List[str]]
            Dictionary mapping level names to level values.
            For example::

                {
                    "method": ("a", "b"),
                    "activation_fn": ("relu", "sigmoid")
                }

        level_order : Tuple[str, ...] | List[str]
            Order of the levels in the result tree. This should contain all
            the keys in ``levels``.
        """
        self.shape = shape
        self.method_names = method_names
        self.levels = levels
        self.level_order = level_order
        self.tree = RandomAccessNDArrayTree(levels, shape)

    def add(self, batch_result: BatchResult):
        """
        Adds a BatchResult to the result object.

        Parameters
        ----------
        batch_result : BatchResult
            BatchResult to add to the result object.
        """
        indices_dict = {
            method_name: np.array(
                [
                    i
                    for i, name in enumerate(batch_result.method_names)
                    if name == method_name
                ]
            )
            for method_name in set(batch_result.method_names)
        }
        target_indices = batch_result.indices.detach().cpu().numpy()
        level_order = list(self.level_order)
        level_order.remove("method")
        self.tree.write_dict_split(
            indices_dict,
            target_indices=target_indices,
            split_level="method",
            data=batch_result.results,
            level_order=level_order,
        )

    def save(self, path: str, format: str) -> None:
        """
        Save the result to an HDF5 file or a nested directory of CSV files.

        Parameters
        ----------
        path : str
            Path to the file.
        format : str
            Format to save the result in. Options: hdf5, dir.
            If hdf5, the full result is stored in a single HDF5 file.
            If csv, the result is stored in a nested directory of CSV files.
        """
        if format == "hdf5":
            with h5py.File(path, mode="x") as fp:
                fp.attrs["type"] = self.__class__.__name__
                self.tree.save_to_hdf(fp)
        elif format == "csv":
            os.makedirs(path, exist_ok=True)
            with open(os.path.join(path, "metadata.yaml"), "w") as fp:
                yaml.dump({"type": self.__class__.__name__}, fp)
            self.tree.save_to_dir(path)
        else:
            raise ValueError("Invalid format: {}".format(format))

    @classmethod
    def _load_tree(cls, path: str, format="hdf5") -> RandomAccessNDArrayTree:
        if format == "hdf5":
            with h5py.File(path, "r") as fp:
                tree = RandomAccessNDArrayTree.load_from_hdf(fp)
        elif format == "csv":
            tree = RandomAccessNDArrayTree.load_from_dir(path)
        else:
            raise ValueError("Invalid format", format)
        return tree

    @classmethod
    def load(cls, path: str) -> "MetricResult":
        """
        Load a result from an HDF5 file or a directory of CSV files.
        The format is inferred from the path: if the path is a directory,
        the result is loaded from a directory of CSV files, otherwise
        the result is loaded from an HDF5 file.

        The specific subclass of MetricResult is inferred from the metadata
        stored in the file or directory, and the appropriate load method
        is called.

        Parameters
        ----------
        path : str
            Path to the file or directory.

        Returns
        -------
        MetricResult
            The loaded result.
        """
        # If the path is a directory, load from directory of CSV files
        if os.path.isdir(path):
            with open(os.path.join(path, "metadata.yaml")) as fp:
                metadata = yaml.safe_load(fp)
            class_name = metadata["type"]
            class_obj = getattr(result, class_name)
            return class_obj._load(path, format="csv")
        # Otherwise, load from HDF5 file
        else:
            with h5py.File(path, "r") as fp:
                class_name = fp.attrs["type"]
                if not isinstance(class_name, str):
                    raise ValueError("Invalid type in HDF5 file")
                class_obj = getattr(result, class_name)
                return class_obj._load(path, format="hdf5")

    @abstractmethod
    def get_df(self, *args, **kwargs) -> Tuple[pd.DataFrame, bool]:
        """
        Retrieve a dataframe from the result object for some given arguments,
        along with a boolean indicating if higher is better.
        These arguments depend on the specific metric.
        """
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def _load(cls, path: str, format="hdf5") -> "MetricResult":
        raise NotImplementedError



import pandas as pd
import seaborn as sns
from sklearn.preprocessing import MinMaxScaler
import numpy as np
from scipy.cluster.hierarchy import linkage
from attribench.plot import Plot
from matplotlib.figure import Figure


class ClusterPlot(Plot):
    """
    Clustermap of the median values of the metrics and methods.
    Allows the user to see which metrics and/or methods behave similarly.

    The plot is shown as a heatmap, with each cell corresponding to the median
    metric value for a given method and metric. The heatmap is clustered using
    hierarchical clustering, with the distance between two methods being the
    correlation between their median metric values. The distance between two
    metrics is computed in the same way. The clustering is performed using
    single linkage.
    """
    def render(self, figsize=(7, 7)) -> Figure:
        """Render the plot.
        TODO add more parameters for font size, title, etc.

        Parameters
        ----------
        figsize : Tuple[int, int], optional
            Figure size, by default (7, 7)
        """
        medians = {}
        for metric_name, (df, higher_is_better) in self.dfs.items():
            medians[metric_name] = (
                df.median(axis=0) if higher_is_better else -df.median(axis=0)
            )

        df = pd.DataFrame.from_dict(medians)
        normalized = MinMaxScaler().fit_transform(df)
        normalized_df = pd.DataFrame(
            normalized, columns=df.columns, index=df.index
        )

        # Manually computing the linkage so that we can set 
        # optimal_ordering to True
        # This should allow for easy comparison between clustermaps
        row_linkage = linkage(
            normalized,
            method="single",
            metric="correlation",
            optimal_ordering=True,
        )
        col_linkage = linkage(
            np.transpose(normalized),
            method="single",
            metric="correlation",
            optimal_ordering=True,
        )
        fig = sns.clustermap(
            normalized_df,
            row_linkage=row_linkage,
            col_linkage=col_linkage,
            figsize=figsize,
        ).fig
        return fig



from typing import Tuple, List
import pandas as pd
from attribench.plot._lib import heatmap
from attribench.plot import Plot
from attribench._stat import wilcoxon_tests
from matplotlib.figure import Figure
import seaborn as sns


class WilcoxonSummaryPlot(Plot):
    """Summary plot for Wilcoxon tests.
    Plots the significance and effect sizes of Wilcoxon tests for each metric
    and each method in a heatmap. Every glyph on the heatmap corresponds to a
    Wilcoxon test between a given method and the random baseline on the given
    metric.
    """
    def render(
        self,
        title: str | None = None,
        figsize: Tuple[int, int] = (20, 20),
        fontsize: int | None = None,
        glyph_scale: int = 1500,
        method_order: List[str] | None =None,
    ) -> Figure:
        """Render the plot.

        Parameters
        ----------
        title : str | None, optional
            Title of the figure, by default None
        figsize : Tuple[int, int], optional
            Figure size, by default (20, 20)
        fontsize : int | None, optional
            Font size of x and y axis ticks, by default None
        glyph_scale : int, optional
            Scale of heatmap glyphs, by default 1500
        method_order : List[str] | None, optional
            Order in which methods should be displayed.
            If None, the order of the keys in the dictionary that was passed to
            the constructor will be used.
            By default None.

        Returns
        -------
        Figure
            The rendered Matplotlib figure.
        """
        pvalues, effect_sizes = {}, {}
        for metric_name, (df, higher_is_better) in self.dfs.items():
            mes, mpv = wilcoxon_tests(df, higher_is_better)
            effect_sizes[metric_name] = mes
            pvalues[metric_name] = mpv
        pvalues = pd.DataFrame(pvalues)
        effect_sizes = pd.DataFrame(effect_sizes).abs()
        effect_sizes[pvalues > 0.01] = 0

        # Normalize each column of the effect sizes dataframe
        effect_sizes = effect_sizes / effect_sizes.max()
        effect_sizes = effect_sizes.fillna(0)

        effect_sizes = pd.melt(effect_sizes.reset_index(), id_vars=["index"])
        effect_sizes.columns = ["method", "metric", "value"]

        return heatmap(
            x=effect_sizes["method"],
            y=effect_sizes["metric"],
            size=effect_sizes["value"],
            color=effect_sizes["value"],
            palette=sns.color_palette("Greens", n_colors=256),
            figsize=figsize,
            glyph_scale=glyph_scale,
            fontsize=fontsize,
            title=title,
            cbar=False,
            x_labels=method_order,
        )



import numpy as np
from attribench._segmentation import segment_samples, isin
import torch


class PerturbationGenerator:
    def __init__(self):
        self.samples: torch.Tensor
        self.rng = np.random.default_rng()

    def set_samples(self, samples):
        self.samples = samples

    def generate_perturbation(self):
        if self.samples is None:
            raise ValueError(
                "Base samples must be set to generate perturbations"
            )
        return self._generate_perturbation_vectors()

    def _generate_perturbation_vectors(self):
        raise NotImplementedError


class NoisyBaselinePerturbationGenerator(PerturbationGenerator):
    def __init__(self, sd):
        super().__init__()
        self.sd = sd

    # perturbation_size is stdev of noise
    def _generate_perturbation_vectors(self):
        # I = x - (x_0 + \epsilon) where x_0 = 0
        return (
            self.samples
            - torch.randn(*self.samples.shape, device=self.samples.device)
            * self.sd
        )


class GaussianPerturbationGenerator(PerturbationGenerator):
    def __init__(self, sd):
        super().__init__()
        self.sd = sd

    # perturbation_size is stdev of noise
    def _generate_perturbation_vectors(self):
        # I \sim \mathcal{N}(0, self.sd)
        return (
            torch.randn(*self.samples.shape, device=self.samples.device)
            * self.sd
        )


class SquarePerturbationGenerator(PerturbationGenerator):
    def __init__(self, square_size):
        super().__init__()
        self.square_size = square_size

    # perturbation_size is (square height)/(image height)
    def _generate_perturbation_vectors(self):
        height = self.samples.shape[2]
        width = self.samples.shape[3]
        x_loc = self.rng.integers(0, width - self.square_size, size=1).item()
        y_loc = self.rng.integers(0, height - self.square_size, size=1).item()
        perturbation_mask = torch.zeros(
            self.samples.shape, device=self.samples.device
        )
        perturbation_mask[
            :,
            :,
            x_loc : x_loc + self.square_size,
            y_loc : y_loc + self.square_size,
        ] = 1
        perturbation_vector = self.samples * perturbation_mask
        return perturbation_vector


class SegmentRemovalPerturbationGenerator(PerturbationGenerator):
    # perturbation size is number of segments
    def __init__(self, num_segments):
        super().__init__()
        self.num_segments = num_segments

    def set_samples(self, samples: torch.Tensor):
        self.samples = samples
        segmented_images = segment_samples(samples.cpu().numpy())
        self.segmented_images = torch.tensor(
            segmented_images, device=samples.device
        )
        self.segments = [
            np.unique(segmented_images[i, ...])
            for i in range(samples.shape[0])
        ]
        self.rng = np.random.default_rng()

    def _generate_perturbation_vectors(self):
        perturbation_vectors = []
        # Select segments to mask for each sample
        segments_to_mask = torch.tensor(
            np.stack(
                [
                    self.rng.choice(
                        self.segments[i], size=self.num_segments, replace=False
                    )
                    for i in range(self.samples.shape[0])
                ]
            ),
            device=self.samples.device,
        )
        for i in range(self.samples.shape[0]):
            seg_sample = self.segmented_images[i, ...]
            sample = self.samples[i, ...]
            # Create boolean mask of pixels that need to be removed
            to_remove = isin(seg_sample, segments_to_mask[i, ...])
            # Create perturbation vector by multiplying mask with image
            perturbation_vector = sample * to_remove
            perturbation_vectors.append(perturbation_vector)
        return torch.stack(perturbation_vectors, dim=0)



from attribench.distributed._worker import WorkerConfig
import torch
from attribench.data.attributions_dataset._attributions_dataset import (
    GroupedAttributionsDataset,
)
from .._metric_worker import GroupedMetricWorker, WorkerConfig
from typing import Callable, Dict
from torch import nn
from attribench.functional.metrics._parameter_randomization import (
    _parameter_randomization_batch,
)
from attribench._method_factory import MethodFactory
from attribench._model_factory import ModelFactory


class ParameterRandomizationWorker(GroupedMetricWorker):
    def __init__(
        self,
        worker_config: WorkerConfig,
        model_factory: ModelFactory,
        dataset: GroupedAttributionsDataset,
        batch_size: int,
        method_factory: MethodFactory,
        agg_fn: Callable[
            [
                torch.Tensor,
                int,
            ],
            torch.Tensor,
        ]
        | None = None,
        agg_dim: int | None = None,
    ):
        super().__init__(worker_config, model_factory, dataset, batch_size)
        self.method_factory = method_factory
        self.agg_fn = agg_fn
        self.agg_dim = agg_dim

    def setup(self):
        self.randomized_model = self._get_model()
        self.method_dict_rand = self.method_factory(self.randomized_model)

    def process_batch(
        self,
        batch_x: torch.Tensor,
        batch_y: torch.Tensor,
        batch_attr: Dict[str, torch.Tensor],
    ):
        return _parameter_randomization_batch(
            batch_x,
            batch_y,
            batch_attr,
            self.method_dict_rand,
            self.device,
            self.agg_fn,
            self.agg_dim,
        )



import numpy as np
import torch

from attribench._segmentation import segment_samples
from attribench.masking import Masker
from attribench.masking.image import ImageMasker


class MaskingDataset:
    def __init__(self, mode: str, start: float, stop: float, num_steps: int):
        if mode not in ("morf", "lerf"):
            raise ValueError("Mode must be morf or lerf")
        if not ((0.0 <= start <= 1.0) and (0.0 <= stop <= 1.0)):
            raise ValueError("Start and stop must be between 0 and 1")
        self.mode = mode
        self.start = start
        self.stop = stop
        self.num_steps = num_steps

    def __len__(self):
        raise NotImplementedError

    def __getitem__(self, item):
        raise NotImplementedError


class DeletionDataset(MaskingDataset):
    def __init__(
        self,
        mode: str,
        start: float,
        stop: float,
        num_steps: int,
        samples: torch.Tensor,
        attrs: torch.Tensor,
        masker: Masker,
    ):
        super().__init__(mode, start, stop, num_steps)
        self.samples = samples
        self.masker = masker
        self.masker.set_batch(samples, attrs)

        total_features = self.masker.get_num_features()
        self.mask_range = list(
            (np.linspace(start, stop, num_steps) * total_features).astype(int)
        )

    def __len__(self):
        return len(self.mask_range)

    def __getitem__(self, item):
        if item >= len(self.mask_range):
            raise StopIteration
        num_to_mask = self.mask_range[item]
        masked_samples = (
            self.masker.mask_top(num_to_mask)
            if self.mode == "morf"
            else self.masker.mask_bot(num_to_mask)
        )
        return masked_samples


class IrofDataset(MaskingDataset):
    def __init__(
        self,
        mode: str,
        start: float,
        stop: float,
        num_steps: int,
        samples: torch.Tensor,
        masker: ImageMasker,
    ):
        super().__init__(mode, start, stop, num_steps)
        self.samples = samples
        self.masker = masker
        self.segmented_images = torch.tensor(
            segment_samples(samples.cpu().numpy()), device=samples.device
        )

    def __len__(self):
        return self.num_steps

    def __getitem__(self, item):
        if item >= self.num_steps:
            raise StopIteration
        to_mask = self.start + (item / (self.num_steps - 1)) * (
            self.stop - self.start
        )
        return (
            self.masker.mask_top(to_mask)
            if self.mode == "morf"
            else self.masker.mask_bot(to_mask)
        )

    def set_attrs(self, attrs: torch.Tensor):
        self.masker.set_batch(self.samples, attrs, self.segmented_images)



from attribench.masking import ImageMasker
import torch


class ConstantMasker(ImageMasker):
    def __init__(self, feature_level, mask_value=0.0):
        super().__init__(feature_level)
        self.mask_value = mask_value

    def _initialize_baselines(self, samples: torch.tensor):
        self.baseline = (
            torch.ones(samples.shape, device=samples.device) * self.mask_value
        )



from typing import Callable, List, Dict

import torch

from attribench._activation_fns import ACTIVATION_FNS
from ._dataset import MaskingDataset


def get_predictions(
    masking_dataset: MaskingDataset,
    labels: torch.Tensor,
    model: Callable,
    activation_fns: List[str],
) -> Dict[str, torch.Tensor]:
    preds = {fn: [] for fn in activation_fns}
    for i, batch in enumerate(iter(masking_dataset)):
        with torch.no_grad():
            predictions = model(batch)
        for fn in activation_fns:
            preds[fn].append(
                ACTIVATION_FNS[fn](predictions)
                .gather(dim=1, index=labels.unsqueeze(-1))
                .cpu()
            )
    preds_cat = {}
    for afn in activation_fns:
        preds_cat[afn] = torch.cat(
            preds[afn], dim=1
        ).cpu()  # [batch_size, len(mask_range)]
    return preds_cat



from ._image_masker import ImageMasker
from ._blurring_image_masker import BlurringImageMasker
from ._constant_image_masker import ConstantImageMasker
from ._random_image_masker import RandomImageMasker
from ._sample_average_image_masker import SampleAverageImageMasker


from typing import Tuple, Optional, List
import pandas as pd
from ._grouped_metric_result import GroupedMetricResult


class ParameterRandomizationResult(GroupedMetricResult):
    """Represents results from running the Parameter Randomization metric.
    """
    def __init__(self, method_names: List[str], num_samples: int):
        """
        Parameters
        ----------
        method_names : List[str]
            Names of attribution methods tested by Parameter Randomization.
        num_samples : int
            Number of samples on which Parameter Randomization was run.
        """
        levels = {"method": method_names}
        level_order = ["method"]
        shape = [num_samples]
        super().__init__(method_names, shape, levels, level_order)

    @classmethod
    def _load(cls, path: str, format="hdf5") -> "ParameterRandomizationResult":
        tree = cls._load_tree(path, format)
        res = ParameterRandomizationResult(tree.levels["method"], tree.shape[0])
        res.tree = tree
        return res

    def get_df(
        self, methods: Optional[List[str]] = None
    ) -> Tuple[pd.DataFrame, bool]:
        """Retrieves a dataframe from the result. The dataframe contains a row
        for each method and a column for each sample. Each value is the
        result value for the given method on the given sample.

        Parameters
        ----------
        methods : Optional[List[str]], optional
            the methods to include. If None, includes all methods.
            Defaults to None.

        Returns
        -------
        Tuple[pd.DataFrame, bool]
            Dataframe containing results,
            and boolean indicating if higher is better.
        """
        methods = methods if methods is not None else self.method_names
        df_dict = {}
        for method in methods:
            df_dict[method] = self.tree.get(method=method)
        return pd.DataFrame.from_dict(df_dict), False



from abc import abstractmethod
from .._distributed_sampler import DistributedSampler
from .._worker import Worker, WorkerConfig
from .._message import PartialResultMessage
from attribench.result._batch_result import BatchResult
from attribench.result._grouped_batch_result import GroupedBatchResult
from typing import Callable, Dict
from attribench.data.attributions_dataset._attributions_dataset import (
    GroupedAttributionsDataset,
    AttributionsDataset,
)
from attribench.data import IndexDataset
from torch import nn
from torch.utils.data import DataLoader
import torch


class MetricWorker(Worker):
    def __init__(
        self,
        worker_config: WorkerConfig,
        model_factory: Callable[[], nn.Module],
        dataset: IndexDataset,
        batch_size: int,
    ):
        super().__init__(worker_config)
        self.batch_size = batch_size
        self.dataset = dataset
        self.model_factory = model_factory

        sampler = DistributedSampler(
            self.dataset,
            self.worker_config.world_size,
            self.worker_config.rank,
            shuffle=False,
        )
        self.dataloader = DataLoader(
            self.dataset,
            sampler=sampler,
            batch_size=self.batch_size,
            num_workers=4,
        )
        self.device = torch.device(self.worker_config.rank)

    def _get_model(self) -> nn.Module:
        model = self.model_factory()
        model.to(self.device)
        return model

    def setup(self):
        self.model = self._get_model()

    def work(self):
        self.setup()

        for (
            batch_indices,
            batch_x,
            batch_y,
            batch_attr,
            method_names,
        ) in self.dataloader:
            batch_x = batch_x.to(self.device)
            batch_y = batch_y.to(self.device)

            batch_result = self.process_batch(
                batch_x,
                batch_y,
                batch_attr,
            )
            self.worker_config.send_result(
                PartialResultMessage(
                    self.worker_config.rank,
                    BatchResult(batch_indices, batch_result, method_names),
                )
            )

    @abstractmethod
    def process_batch(
        self,
        batch_x: torch.Tensor,
        batch_y: torch.Tensor,
        batch_attr: torch.Tensor,
    ):
        raise NotImplementedError


class GroupedMetricWorker(MetricWorker):
    def __init__(
        self,
        worker_config: WorkerConfig,
        model_factory: Callable[[], nn.Module],
        dataset: GroupedAttributionsDataset,
        batch_size: int,
    ):
        super().__init__(
            worker_config,
            model_factory,
            dataset,
            batch_size,
        )

    def work(self):
        self.setup()

        for (
            batch_indices,
            batch_x,
            batch_y,
            batch_attr,
        ) in self.dataloader:
            batch_x = batch_x.to(self.device)
            batch_y = batch_y.to(self.device)

            batch_result = self.process_batch(
                batch_x,
                batch_y,
                batch_attr,
            )
            self.worker_config.send_result(
                PartialResultMessage(
                    self.worker_config.rank,
                    GroupedBatchResult(batch_indices, batch_result),
                )
            )

    @abstractmethod
    def process_batch(
        self,
        batch_x: torch.Tensor,
        batch_y: torch.Tensor,
        batch_attr: Dict[str, torch.Tensor],
    ):
        raise NotImplementedError



import pandas as pd
from tqdm import tqdm
import seaborn as sns
import matplotlib.pyplot as plt
from attribench.plot import Plot
from matplotlib.figure import Figure


class ConvergencePlot(Plot):
    """ Line plot of the median values of a given metric vs the number of samples.
    Error bars are computed using bootstrapping.
    Allows the user to inspect if metric values have converged,
    i.e. if the benchmark has been run on enough samples.
    """

    def __init__(self, df: pd.DataFrame):
        """
        Parameters
        ----------
        df : pd.DataFrame
            Pandas dataframe containing the metric values.
            The columns are the names of the methods.
        """
        super().__init__({})
        self.df = df

    def render(
        self,
        title: str | None = None,
        figsize=(20, 20),
        fontsize=None,
        bs_samples=1000,
        interval=5,
    ) -> Figure:
        """Render the plot.

        Parameters
        ----------
        title : str | None, optional
            Title of the figure, by default None
        figsize : Tuple[int, int], optional
            Figure size, by default (20, 20)
        fontsize : int | None, optional
            Font size of x and y axis ticks, by default None
        bs_samples : int, optional
            Number of bootstrap samples for estimating the median value of the
            metric using a given sample size. By default 1000
        interval : int, optional
            Interval between sample sizes, by default 5

        Returns
        -------
        Figure
            The rendered Matplotlib figure.
        """
        all_medians = []
        for bs_size in tqdm(range(2, self.df.shape[0], interval)):
            medians = []
            for _ in range(bs_samples):
                sample = self.df.sample(n=bs_size, replace=True)
                medians.append(sample.median(axis=0))  # median for each column
            medians = pd.DataFrame(medians)

            medians = pd.melt(medians, var_name="method")
            medians["bs_size"] = bs_size

            all_medians.append(medians)
        all_medians = pd.concat(all_medians)
        fig, ax = plt.subplots(figsize=figsize)
        sns.lineplot(
            data=all_medians,
            x="bs_size",
            y="value",
            hue="method",
            estimator="median",
            errorbar="ci",
            ax=ax,
        )
        ax.set_xticklabels(ax.get_xticklabels(), size=fontsize)
        ax.set_yticklabels(ax.get_yticklabels(), size=fontsize)
        plt.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.0)
        if title is not None:
            ax.set_title(title)
        return fig



from ._plot import Plot
from ._wilcoxon_summary_plot import WilcoxonSummaryPlot
from ._correlations import InterMetricCorrelationPlot, InterMethodCorrelationPlot
from ._convergence_plot import ConvergencePlot
from ._krippendorff_alpha_plot import KrippendorffAlphaPlot
from ._cluster_plot import ClusterPlot
from ._mad_ratio_plot import MADRatioPlot
from ._wilcoxon_barplot import WilcoxonBarPlot
from ._cles_plot import CLESPlot


from typing import Callable, Union, Tuple, List, Optional, Mapping
import torch
from torch import multiprocessing as mp
from torch.utils.data import DataLoader
from attribench.distributed._message import PartialResultMessage

from attribench.masking import ImageMasker, Masker
from attribench._model_factory import ModelFactory
from ..._worker import Worker
from ..._distributed_sampler import DistributedSampler
from ..._message import PartialResultMessage
from attribench.result._batch_result import BatchResult
from ._deletion import Deletion
from ._deletion_worker import DeletionWorker
from attribench.data import AttributionsDataset
from attribench.functional.metrics._irof import _irof_batch

from multiprocessing.synchronize import Event


class IrofWorker(DeletionWorker):
    def __init__(
        self,
        result_queue: mp.Queue,
        rank: int,
        world_size: int,
        all_processes_done: Event,
        model_factory: ModelFactory,
        dataset: AttributionsDataset,
        batch_size: int,
        maskers: Mapping[str, ImageMasker],
        activation_fns: List[str],
        mode: str = "morf",
        start: float = 0,
        stop: float = 1,
        num_steps: int = 100,
        result_handler: Callable[[PartialResultMessage], None] | None = None,
    ):
        super().__init__(
            result_queue,
            rank,
            world_size,
            all_processes_done,
            model_factory,
            dataset,
            batch_size,
            maskers,
            activation_fns,
            mode,
            start,
            stop,
            num_steps,
            result_handler,
        )
        self.maskers = maskers

    def work(self):
        sampler = DistributedSampler(
            self.dataset, self.world_size, self.rank, shuffle=False
        )
        dataloader = DataLoader(
            self.dataset,
            sampler=sampler,
            batch_size=self.batch_size,
            num_workers=4,
        )
        device = torch.device(self.rank)
        model = self.model_factory()
        model.to(device)

        for (
            batch_indices,
            batch_x,
            batch_y,
            batch_attr,
            method_names,
        ) in dataloader:
            batch_x = batch_x.to(device)
            batch_y = batch_y.to(device)
            batch_result = _irof_batch(
                batch_x,
                batch_y,
                model,
                batch_attr.numpy(),
                self.maskers,
                self.activation_fns,
                self.mode,
                self.start,
                self.stop,
                self.num_steps,
            )
            self.send_result(
                PartialResultMessage(
                    self.rank,
                    BatchResult(batch_indices, batch_result, method_names),
                )
            )


class Irof(Deletion):
    def __init__(
        self,
        model_factory: ModelFactory,
        dataset: AttributionsDataset,
        batch_size: int,
        maskers: Mapping[str, ImageMasker],
        activation_fns: Union[List[str], str],
        mode: str = "morf",
        start: float = 0.0,
        stop: float = 1.0,
        num_steps: int = 100,
        address="localhost",
        port="12355",
        devices: Optional[Tuple] = None,
    ):
        """Compute the IROF metric for a given `AttributionsDataset` and model
        using multiple processes.

        IROF starts segmenting the input image using SLIC. Then, it iteratively
        masks out the top (Most Relevant First, or MoRF) or bottom (Least Relevant
        First, or LeRF) segments and computes the confidence of the model on the
        masked samples. The relevance of a segment is computed as the average
        relevance of the features in the segment.

        This results in a curve of confidence vs. number of segments masked. The
        area under (or equivalently over) this curve is the IROF metric.

        `start`, `stop`, and `num_steps` are used to determine the range of segments
        to mask. The range is determined by `start` and `stop` as a percentage of
        the total number of segments. `num_steps` is the number of steps to take
        between `start` and `stop`.

        The IROF metric is computed for each masker in `maskers` and for each
        activation function in `activation_fns`. The number of processes is
        determined by the number of devices. If `devices` is None, then all
        available devices are used. Samples are distributed evenly across the
        processes.

        Parameters
        ----------
        model_factory : ModelFactory
            ModelFactory instance or callable that returns a model.
            Used to create a model for each subprocess.
        dataset : AttributionsDataset
            Dataset containing the samples and attributions to compute
            IROF on.
        batch_size : int
            The batch size to use when computing the metric.
        maskers : Dict[str, Masker]
            Dictionary of maskers to use for computing the metric.
        activation_fns : Union[List[str], str]
            List of activation functions to use for computing the metric.
            If a single string is given, it is converted to a single-element
            list.
        mode : str, optional
            Mode to use for computing the metric. Either "morf" or "lerf".
            Default: "morf"
        start : float, optional
            Relative start of the range of features to mask. Must be between 0 and 1.
            Default: 0.0
        stop : float, optional
            Relative end of the range of features to mask. Must be between 0 and 1.
            Default: 1.0
        num_steps : int, optional
            Number of steps to use for the range of features to mask.
            Default: 100
        address : str, optional
            Address to use for the multiprocessing connection.
            Default: "localhost"
        port : str, optional
            Port to use for the multiprocessing connection.
            Default: "12355"
        devices : Optional[Tuple], optional
            Devices to use. If None, then all available devices are used.
            Default: None
        """
        super().__init__(
            model_factory,
            dataset,
            batch_size,
            maskers,
            activation_fns,
            mode,
            start,
            stop,
            num_steps,
            address,
            port,
            devices,
        )
        self.maskers = maskers

    def _create_worker(
        self, queue: mp.Queue, rank: int, all_processes_done: Event
    ) -> Worker:
        return IrofWorker(
            queue,
            rank,
            self.world_size,
            all_processes_done,
            self.model_factory,
            self.dataset,
            self.batch_size,
            self.maskers,
            self.activation_fns,
            self.mode,
            self._start,
            self.stop,
            self.num_steps,
            self._handle_result if self.world_size == 1 else None,
        )



import torch
from .._metric_worker import MetricWorker, WorkerConfig
from typing import Mapping, List
from attribench.masking import Masker
from attribench.data import AttributionsDataset
from attribench.functional.metrics.deletion._deletion import _deletion_batch
from attribench._model_factory import ModelFactory


class DeletionWorker(MetricWorker):
    def __init__(
        self,
        worker_config: WorkerConfig,
        model_factory: ModelFactory,
        dataset: AttributionsDataset,
        batch_size: int,
        maskers: Mapping[str, Masker],
        activation_fns: List[str],
        mode: str = "morf",
        start: float = 0.0,
        stop: float = 1.0,
        num_steps: int = 100,
    ):
        super().__init__(worker_config, model_factory, dataset, batch_size)
        self.maskers = maskers
        self.activation_fns = activation_fns
        self.mode = mode
        self.start = start
        self.stop = stop
        self.num_steps = num_steps

    def process_batch(
        self,
        batch_x: torch.Tensor,
        batch_y: torch.Tensor,
        batch_attr: torch.Tensor,
    ):
        return _deletion_batch(
            batch_x,
            batch_y,
            self.model,
            batch_attr,
            self.maskers,
            self.activation_fns,
            self.mode,
            self.start,
            self.stop,
            self.num_steps,
        )


from attribench.masking import ImageMasker
import torch


class RandomMasker(ImageMasker):
    def __init__(self, feature_level, std=1, num_samples=1):
        super().__init__(feature_level)
        self.std = std
        self.num_samples = num_samples

    def _initialize_baselines(self, samples: torch.tensor):
        self.baseline = (
            torch.randn(*samples.shape, device=samples.device) * self.std
        )



from torch import nn
from torch.utils.data import Dataset, DataLoader
from attribench.data import HDF5DatasetWriter
import torch
from typing import Optional, Tuple


def _select_samples_batch(
    batch_x: torch.Tensor,
    batch_y: torch.Tensor,
    model: nn.Module,
    device: torch.device,
):
    """Returns the correctly classified samples and their labels.

    Parameters
    ----------
    batch_x : torch.Tensor
    batch_y : torch.Tensor
    model : nn.Module
    device : torch.device

    Returns
    -------
    Tuple[torch.Tensor, torch.Tensor]
        The correctly classified samples and their labels.
    """
    batch_x = batch_x.to(device)
    batch_y = batch_y.to(device)
    with torch.no_grad():
        output = torch.argmax(model(batch_x), dim=1)
    correct_samples = batch_x[output == batch_y, ...]
    correct_labels = batch_y[output == batch_y]
    return correct_samples, correct_labels


def select_samples(
    model: nn.Module,
    dataset: Dataset,
    num_samples: int,
    batch_size: int,
    writer: Optional[HDF5DatasetWriter] = None,
    device: Optional[torch.device] = None,
) -> Optional[Tuple[torch.Tensor, torch.Tensor]]:
    """Select correctly classified samples from a dataset and optionally
    write them to a HDF5 file. If the `writer` is `None`, the
    samples and labels are simply returned. Otherwise, the samples and
    labels are written to the HDF5 file and `None` is returned.

    TODO this function should just return the samples and labels. Use the
    distributed class to write the samples and labels to a file.

    Parameters
    ----------
    model : nn.Module
        Model to use for classification.
    dataset : Dataset
        Torch Dataset containing the samples and labels.
    writer : HDF5DatasetWriter
        Writer to write the samples and labels to, by default None.
    num_samples : int
        Number of correctly classified samples to select.
    batch_size : int
        Batch size to use for the dataloader.
    device : Optional[torch.device], optional
        Device to use, by default None.

    Returns
    -------
    Tuple[torch.Tensor, torch.Tensor] | None
        If `writer` is `None`, a tuple containing the correctly classified
        samples and their labels. Otherwise, `None`.
    """
    if device is None:
        device = torch.device("cpu")

    model.to(device)
    model.eval()

    dataloader = DataLoader(
        dataset, batch_size=batch_size, num_workers=4, pin_memory=True
    )

    samples_count = 0
    all_correct_samples, all_correct_labels = [], []
    for batch_x, batch_y in dataloader:
        correct_samples, correct_labels = _select_samples_batch(
            batch_x, batch_y, model, device
        )
        if len(correct_samples) > 0:
            if writer is None:
                all_correct_samples.append(correct_samples)
                all_correct_labels.append(correct_labels)
            else:
                writer.write(
                    correct_samples.cpu().numpy(), correct_labels.cpu().numpy()
                )
            samples_count += len(correct_samples)
        if samples_count >= num_samples:
            break
    if writer is None:
        return torch.cat(all_correct_samples), torch.cat(all_correct_labels)



from typing import Union, Tuple, Optional, List, Mapping
from attribench.masking import Masker
from ._deletion_worker import DeletionWorker
from ..._worker import WorkerConfig
from attribench.result import DeletionResult
from .._metric import Metric
from attribench.data import AttributionsDataset
from attribench import ModelFactory


class Deletion(Metric):
    """Compute the Deletion metric for a given :class:`~attribench.data.AttributionsDataset` and model
    using multiple processes.

    Deletion is computed by iteratively masking the top (Most Relevant First,
    or MoRF) or bottom (Least Relevant First, or LeRF) features of
    the input samples and computing the confidence of the model on the masked
    samples.

    This results in a curve of confidence vs. number of features masked. The
    area under (or equivalently over) this curve is the Deletion metric.

    `start`, `stop`, and `num_steps` are used to determine the range of features
    to mask. The range is determined by `start` and `stop` as a percentage of
    the total number of features. `num_steps` is the number of steps to take
    between `start` and `stop`.

    The Deletion metric is computed for each masker in `maskers` and for each
    activation function in `activation_fns`. The number of processes is
    determined by the number of devices. If `devices` is None, then all
    available devices are used. Samples are distributed evenly across the
    processes.
    """

    def __init__(
        self,
        model_factory: ModelFactory,
        attributions_dataset: AttributionsDataset,
        batch_size: int,
        maskers: Mapping[str, Masker],
        activation_fns: Union[List[str], str],
        mode: str = "morf",
        start: float = 0.0,
        stop: float = 1.0,
        num_steps: int = 100,
        address="localhost",
        port="12355",
        devices: Optional[Tuple] = None,
    ):
        """
        Parameters
        ----------
        model_factory : ModelFactory
            ModelFactory instance or callable that returns a model.
            Used to create a model for each subprocess.
        attributions_dataset : AttributionsDataset
            Dataset containing the samples and attributions to compute
            Deletion on.
        batch_size : int
            The batch size to use when computing the metric.
        maskers : Dict[str, Masker]
            Dictionary of maskers to use for computing the metric.
        activation_fns : Union[List[str], str]
            List of activation functions to use for computing the metric.
            If a single string is given, it is converted to a single-element
            list.
        mode : str, optional
            Mode to use for computing the metric. Either "morf" or "lerf".
            Default: "morf"
        start : float, optional
            Relative start of the range of features to mask. Must be between 0 and 1.
            Default: 0.0
        stop : float, optional
            Relative end of the range of features to mask. Must be between 0 and 1.
            Default: 1.0
        num_steps : int, optional
            Number of steps to use for the range of features to mask.
            Default: 100
        address : str, optional
            Address to use for the multiprocessing connection.
            Default: "localhost"
        port : str, optional
            Port to use for the multiprocessing connection.
            Default: "12355"
        devices : Optional[Tuple], optional
            Devices to use. If None, then all available devices are used.
            Default: None
        """
        super().__init__(
            model_factory, attributions_dataset, batch_size, address, port, devices
        )
        self.num_steps = num_steps
        self.stop = stop
        self._start = start
        self.mode = mode
        if isinstance(activation_fns, str):
            activation_fns = [activation_fns]
        self.activation_fns: List[str] = activation_fns
        self.maskers = maskers
        self._result = DeletionResult(
            attributions_dataset.method_names,
            list(maskers.keys()),
            self.activation_fns,
            mode,
            attributions_dataset.num_samples,
            num_steps,
        )
        self.dataset = attributions_dataset

    def _create_worker(self, worker_config: WorkerConfig) -> DeletionWorker:
        return DeletionWorker(
            worker_config,
            self.model_factory,
            self.dataset,
            self.batch_size,
            self.maskers,
            self.activation_fns,
            self.mode,
            self._start,
            self.stop,
            self.num_steps,
        )
